from typing import Any

__all__ = ["UnsupportedTransformerArgException"]


class UnsupportedTransformerArgException(Exception):
    def __init__(self, arg: Any):
        super().__init__(f"Unsupported transformer argument: {arg}")



from tests.lib.exceptions import NumberIsEven, \
    NumberIsOdd, \
    NumberLessThanOrEquals10, \
    NumbersNotEqual


def is_even(num: int):
    if num % 2 != 0:
        raise NumberIsOdd()


def is_odd(num: int):
    if num % 2 == 0:
        raise NumberIsEven()


def is_greater_than_10(num: float):
    if num <= 10:
        raise NumberLessThanOrEquals10()


def same_value(data: float, output: float):
    if data != output:
        raise NumbersNotEqual()


def same_value_int(incoming: int, outcome: int):
    if incoming != outcome:
        raise NumbersNotEqual()




import asyncio
import types
from inspect import Signature
from types import GenericAlias
from typing import TypeVar, Any, cast

from gloe.async_transformer import AsyncTransformer
from gloe.base_transformer import BaseTransformer
from gloe.transformers import Transformer
from gloe._utils import _match_types, _specify_types, awaitify
from gloe.exceptions import UnsupportedTransformerArgException

_In = TypeVar("_In")
_Out = TypeVar("_Out")
_NextOut = TypeVar("_NextOut")


def is_transformer(node):
    if type(node) == list or type(node) == tuple:
        return all(is_transformer(n) for n in node)
    return isinstance(node, Transformer)


def is_async_transformer(node):
    return isinstance(node, AsyncTransformer)


def has_any_async_transformer(node: list):
    return any(is_async_transformer(n) for n in node)


def _resolve_new_merge_transformers(
    new_transformer: BaseTransformer, transformer2: BaseTransformer
):
    new_transformer.__class__.__name__ = transformer2.__class__.__name__
    new_transformer._label = transformer2.label
    new_transformer._children = transformer2.children
    new_transformer._invisible = transformer2.invisible
    new_transformer._graph_node_props = transformer2.graph_node_props
    new_transformer._set_previous(transformer2.previous)
    return new_transformer


def _resolve_serial_connection_signatures(
    transformer2: BaseTransformer, generic_vars: dict, signature2: Signature
) -> Signature:
    first_param = list(signature2.parameters.values())[0]
    new_parameter = first_param.replace(
        annotation=_specify_types(transformer2.input_type, generic_vars)
    )
    new_signature = signature2.replace(
        parameters=[new_parameter],
        return_annotation=_specify_types(signature2.return_annotation, generic_vars),
    )
    return new_signature


def _nerge_serial(transformer1, _transformer2):
    if transformer1.previous is None:
        transformer1 = transformer1.copy(regenerate_instance_id=True)

    transformer2 = _transformer2.copy(regenerate_instance_id=True)
    transformer2._set_previous(transformer1)

    signature1: Signature = transformer1.signature()
    signature2: Signature = transformer2.signature()

    input_generic_vars = _match_types(
        transformer2.input_type, signature1.return_annotation
    )
    output_generic_vars = _match_types(
        signature1.return_annotation, transformer2.input_type
    )
    generic_vars = {**input_generic_vars, **output_generic_vars}

    def transformer1_signature(_) -> Signature:
        return signature1.replace(
            return_annotation=_specify_types(signature1.return_annotation, generic_vars)
        )

    setattr(
        transformer1,
        "signature",
        types.MethodType(transformer1_signature, transformer1),
    )

    class BaseNewTransformer:
        def signature(self) -> Signature:
            return _resolve_serial_connection_signatures(
                transformer2, generic_vars, signature2
            )

        def __len__(self):
            return len(transformer1) + len(transformer2)

    new_transformer: BaseTransformer | None = None
    if is_transformer(transformer1) and is_transformer(_transformer2):

        class NewTransformer1(BaseNewTransformer, Transformer[_In, _NextOut]):
            def transform(self, data: _In) -> _NextOut:
                transformer2_call = transformer2.__call__
                transformer1_call = transformer1.__call__
                transformed = transformer2_call(transformer1_call(data))
                return transformed

        new_transformer = NewTransformer1()

    elif is_async_transformer(transformer1) and is_transformer(_transformer2):

        class NewTransformer2(BaseNewTransformer, AsyncTransformer[_In, _NextOut]):
            async def transform_async(self, data: _In) -> _NextOut:
                transformer1_out = await transformer1(data)
                transformed = transformer2(transformer1_out)
                return transformed

        new_transformer = NewTransformer2()

    elif is_async_transformer(transformer1) and is_async_transformer(transformer2):

        class NewTransformer3(BaseNewTransformer, AsyncTransformer[_In, _NextOut]):
            async def transform_async(self, data: _In) -> _NextOut:
                transformer1_out = await transformer1(data)
                transformed = await transformer2(transformer1_out)
                return transformed

        new_transformer = NewTransformer3()

    elif is_transformer(transformer1) and is_async_transformer(_transformer2):

        class NewTransformer4(AsyncTransformer[_In, _NextOut]):
            async def transform_async(self, data: _In) -> _NextOut:
                transformer1_out = transformer1(data)
                transformed = await transformer2(transformer1_out)
                return transformed

        new_transformer = NewTransformer4()

    else:
        raise UnsupportedTransformerArgException(transformer2)

    return _resolve_new_merge_transformers(new_transformer, transformer2)


def _merge_diverging(
    incident_transformer,
    *receiving_transformers,
):
    if incident_transformer.previous is None:
        incident_transformer = incident_transformer.copy(regenerate_instance_id=True)

    receiving_transformers = tuple(
        [
            receiving_transformer.copy(regenerate_instance_id=True)
            for receiving_transformer in receiving_transformers
        ]
    )

    for receiving_transformer in receiving_transformers:
        receiving_transformer._set_previous(incident_transformer)

    incident_signature: Signature = incident_transformer.signature()
    receiving_signatures: list[Signature] = []

    for receiving_transformer in receiving_transformers:
        generic_vars = _match_types(
            receiving_transformer.input_type, incident_signature.return_annotation
        )

        receiving_signature = receiving_transformer.signature()
        return_annotation = receiving_signature.return_annotation

        new_return_annotation = _specify_types(return_annotation, generic_vars)

        new_signature = receiving_signature.replace(
            return_annotation=new_return_annotation
        )
        receiving_signatures.append(new_signature)

        def _signature(_) -> Signature:
            return new_signature

        if receiving_transformer._previous == incident_transformer:
            setattr(
                receiving_transformer,
                "signature",
                types.MethodType(_signature, receiving_transformer),
            )

    class BaseNewTransformer:
        def signature(self) -> Signature:
            receiving_signature_returns = [
                r.return_annotation for r in receiving_signatures
            ]
            new_signature = incident_signature.replace(
                return_annotation=GenericAlias(tuple, tuple(receiving_signature_returns))
            )
            return new_signature

        def __len__(self):
            lengths = [len(t) for t in receiving_transformers]
            return sum(lengths) + len(incident_transformer)

    new_transformer = None
    if is_transformer(incident_transformer) and is_transformer(receiving_transformers):

        def split_result(data: _In) -> tuple[Any, ...]:
            intermediate_result = incident_transformer(data)

            outputs = []
            for receiving_transformer in receiving_transformers:
                output = receiving_transformer(intermediate_result)
                outputs.append(output)

            return tuple(outputs)

        class NewTransformer1(BaseNewTransformer, Transformer[_In, tuple[Any, ...]]):
            def transform(self, data: _In) -> tuple[Any, ...]:
                return split_result(data)

        new_transformer = NewTransformer1()

    else:

        async def split_result_async(data: _In) -> tuple[Any, ...]:
            if asyncio.iscoroutinefunction(incident_transformer.__call__):
                intermediate_result = await incident_transformer(data)
            else:
                intermediate_result = incident_transformer(data)

            outputs = []
            for receiving_transformer in receiving_transformers:
                if asyncio.iscoroutinefunction(receiving_transformer.__call__):
                    output = await receiving_transformer(intermediate_result)
                else:
                    output = receiving_transformer(intermediate_result)
                outputs.append(output)

            return tuple(outputs)

        class NewTransformer2(BaseNewTransformer, AsyncTransformer[_In, tuple[Any, ...]]):
            async def transform_async(self, data: _In) -> tuple[Any, ...]:
                return await split_result_async(data)

        new_transformer = NewTransformer2()

    new_transformer._previous = cast(Transformer, receiving_transformers)
    new_transformer.__class__.__name__ = "Converge"
    new_transformer._label = ""
    new_transformer._graph_node_props = {
        "shape": "diamond",
        "width": 0.5,
        "height": 0.5,
    }

    return new_transformer


def _compose_nodes(
    current: BaseTransformer,
    next_node: tuple | BaseTransformer,
):
    if issubclass(type(current), BaseTransformer):
        if issubclass(type(next_node), BaseTransformer):
            return _nerge_serial(current, next_node)  # type: ignore
        elif type(next_node) == tuple:
            is_all_base_transformers = all(
                issubclass(type(next_transformer), BaseTransformer)
                for next_transformer in next_node
            )
            if is_all_base_transformers:
                return _merge_diverging(current, *next_node)  # type: ignore

            unsupported_elem = [
                elem for elem in next_node if not isinstance(elem, BaseTransformer)
            ]
            raise UnsupportedTransformerArgException(unsupported_elem[0])
        else:
            raise UnsupportedTransformerArgException(next_node)
    else:
        raise UnsupportedTransformerArgException(next_node)



import inspect
import warnings
from inspect import Signature
from types import FunctionType
from typing import (
    Callable,
    Concatenate,
    ParamSpec,
    TypeVar,
    cast,
    Awaitable,
    Generic,
)

from gloe.async_transformer import AsyncTransformer
from gloe.transformers import Transformer

__all__ = [
    "transformer",
    "partial_transformer",
    "async_transformer",
    "partial_async_transformer",
]

A = TypeVar("A")
S = TypeVar("S")
S2 = TypeVar("S2")
P1 = ParamSpec("P1")
P2 = ParamSpec("P2")
O = TypeVar("O")


class _PartialTransformer(Generic[A, P1, S]):
    def __init__(self, func: Callable[Concatenate[A, P1], S]):
        self.func = func

    def __call__(self, *args: P1.args, **kwargs: P1.kwargs) -> Transformer[A, S]:
        func = self.func
        func_signature = inspect.signature(func)

        class LambdaTransformer(Transformer[A, S]):
            __doc__ = func.__doc__
            __annotations__ = cast(FunctionType, func).__annotations__

            def signature(self) -> Signature:
                return func_signature

            def transform(self, data: A) -> S:
                return func(data, *args, **kwargs)

        lambda_transformer = LambdaTransformer()
        lambda_transformer.__class__.__name__ = func.__name__
        lambda_transformer._label = func.__name__
        return lambda_transformer


def partial_transformer(
    func: Callable[Concatenate[A, P1], S]
) -> _PartialTransformer[A, P1, S]:
    """
    This decorator let us create partial transformers, which are transformers that
    allow for partial application of their arguments. This capability is particularly
    useful for creating configurable transformer instances where some arguments are preset
    enhancing modularity and reusability in data processing pipelines.

    See Also:
        For further details on partial transformers and their applications, see
        :ref:`partial-transformers`.

    Example:
        Here's how to apply the `@partial_transformer` decorator to create a transformer
        with a pre-applied argument::

            @partial_transformer
            def enrich_data(data: Data, enrichment_type: str) -> Data:
                # Implementation for data enrichment based on the enrichment_type
                ...

            # Instantiate a transformer with the 'enrichment_type' pre-set
            enrich_with_metadata = enrich_data(enrichment_type="metadata")

            # Use the partially applied transformer
            get_enriched_data = get_data >> enrich_with_metadata

    Args:
        func: A callable with one or more arguments. The first argument is of
            type :code:`A`. The subsequent arguments are retained for use during
            transformer instantiation. This callable returns a value of type
            :code:`S`.

    Returns:
        An instance of the :code:`_PartialTransformer`, an internal class utilized within
        Gloe that facilitates partial instantiation of transformers by the user.
        The underlying mechanics of :code:`_PartialTransformer` are managed internally,
        the user just needs to understand its usage.
    """
    return _PartialTransformer(func)


class _PartialAsyncTransformer(Generic[A, P1, S]):
    def __init__(self, func: Callable[Concatenate[A, P1], Awaitable[S]]):
        self.func = func

    def __call__(self, *args: P1.args, **kwargs: P1.kwargs) -> AsyncTransformer[A, S]:
        func = self.func
        func_signature = inspect.signature(func)

        class LambdaTransformer(AsyncTransformer[A, S]):
            __doc__ = func.__doc__
            __annotations__ = cast(FunctionType, func).__annotations__

            def signature(self) -> Signature:
                return func_signature

            async def transform_async(self, data: A) -> S:
                return await func(data, *args, **kwargs)

        lambda_transformer = LambdaTransformer()
        lambda_transformer.__class__.__name__ = func.__name__
        lambda_transformer._label = func.__name__
        return lambda_transformer


def partial_async_transformer(
    func: Callable[Concatenate[A, P1], Awaitable[S]]
) -> _PartialAsyncTransformer[A, P1, S]:
    """
    This decorator enables the creation of partial asynchronous transformers, which are
    transformers capable of partial argument application. Such functionality is invaluable
    for crafting reusable asynchronous transformer instances where certain arguments are
    predetermined, enhancing both modularity and reusability within asynchronous data
    processing flows.

    See Also:
        For additional insights into partial asynchronous transformers and their practical
        applications, consult :ref:`partial-async-transformers`.

    Example:
        Utilize the `@partial_async_transformer` decorator to build a transformer with
        a pre-set argument::

            @partial_async_transformer
            async def load_data(user_id: int, data_type: str) -> Data:
                # Logic for loading data based on user_id and data_type
                ...

            # Instantiate a transformer with 'data_type' predefined
            load_user_data = load_data(data_type="user_profile")

            # Subsequent usage requires only the user_id
            user_data = await load_user_data(user_id=1234)

    Args:
        func: A callable with one or more arguments, the first of which is of type `A`.
            Remaining arguments are preserved for later use during the instantiation of
            the transformer. This callable must asynchronously return a result of type
            `S`, indicating an operation that produces an output of type `S` upon
            completion.

    Returns:
        An instance of the :code:`_PartialAsyncTransformer`, an internally managed class
        within Gloe designed to facilitate the partial instantiation of asynchronous
        transformers. Users are encouraged to understand its application, as the
        underlying mechanics of :code:`_PartialAsyncTransformer` are handled internally.
    """
    return _PartialAsyncTransformer(func)


def transformer(func: Callable[[A], S]) -> Transformer[A, S]:
    """
    Convert a callable to an instance of the Transformer class.

    Example:
        The most common use is as a decorator::

            @transformer
            def filter_subscribed_users(users: list[User]) -> list[User]:
               ...

            subscribed_users = filter_subscribed_users(users_list)

    Args:
        func: A callable that takes a single argument and returns a result. The callable
            should return an instance of the generic type :code:`S` specified.
    Returns:
        An instance of the Transformer class, encapsulating the transformation logic
        defined in the provided callable.
    """
    func_signature = inspect.signature(func)

    if len(func_signature.parameters) > 1:
        warnings.warn(
            "Only one parameter is allowed on Transformers. "
            f"Function '{func.__name__}' has the following signature: {func_signature}. "
            "To pass a complex data, use a complex type like named tuples, "
            "typed dicts, dataclasses or anything else.",
            category=RuntimeWarning,
        )

    class LambdaTransformer(Transformer[A, S]):
        __doc__ = func.__doc__
        __annotations__ = cast(FunctionType, func).__annotations__

        def signature(self) -> Signature:
            return func_signature

        def transform(self, data):
            return func(data)

    lambda_transformer = LambdaTransformer()
    lambda_transformer.__class__.__name__ = func.__name__
    lambda_transformer._label = func.__name__
    return lambda_transformer


def async_transformer(func: Callable[[A], Awaitable[S]]) -> AsyncTransformer[A, S]:
    """
    Convert a callable to an instance of the AsyncTransformer class.

    See Also:
        For more information about this feature, refer to the :ref:`async-transformers`.

    Example:
        The most common use is as a decorator::

            @async_transformer
            async def get_user_by_role(role: str) -> list[User]:
               ...

            await get_user_by_role("admin")

    Args:
        func: A callable that takes a single argument and returns a coroutine.
    Returns:
        Returns an instance of the AsyncTransformer class, representing the built async
        transformer.
    """
    func_signature = inspect.signature(func)

    if len(func_signature.parameters) > 1:
        warnings.warn(
            "Only one parameter is allowed on Transformers. "
            f"Function '{func.__name__}' has the following signature: {func_signature}. "
            "To pass a complex data, use a complex type like named tuples, "
            "typed dicts, dataclasses or anything else.",
            category=RuntimeWarning,
        )

    class LambdaAsyncTransformer(AsyncTransformer[A, S]):
        __doc__ = func.__doc__
        __annotations__ = cast(FunctionType, func).__annotations__

        def signature(self) -> Signature:
            return func_signature

        async def transform_async(self, data):
            return await func(data)

    lambda_transformer = LambdaAsyncTransformer()
    lambda_transformer.__class__.__name__ = func.__name__
    lambda_transformer._label = func.__name__
    return lambda_transformer



import copy
import traceback
import types
import uuid
from abc import abstractmethod, ABC
from inspect import Signature
from typing import TypeVar, overload, cast, Any, Callable, Awaitable

from gloe.base_transformer import (
    TransformerException,
    BaseTransformer,
    PreviousTransformer,
)

__all__ = ["AsyncTransformer"]

_In = TypeVar("_In")
_Out = TypeVar("_Out")
_NextOut = TypeVar("_NextOut")

_Out2 = TypeVar("_Out2")
_Out3 = TypeVar("_Out3")
_Out4 = TypeVar("_Out4")
_Out5 = TypeVar("_Out5")
_Out6 = TypeVar("_Out6")
_Out7 = TypeVar("_Out7")


class AsyncTransformer(BaseTransformer[_In, _Out, "AsyncTransformer"], ABC):
    def __init__(self):
        super().__init__()

        self._graph_node_props: dict[str, Any] = {
            **self._graph_node_props,
            "isAsync": True,
        }
        self.__class__.__annotations__ = self.transform_async.__annotations__

    @abstractmethod
    async def transform_async(self, data: _In) -> _Out:
        """
        Method to perform the transformation asynchronously.

        Args:
            data: the incoming data passed to the transformer during the pipeline execution.

        Return:
            The outcome data, it means, the resulf of the transformation.
        """
        pass

    def signature(self) -> Signature:
        return self._signature(AsyncTransformer)

    def __repr__(self):
        return f"{self.input_annotation} -> ({type(self).__name__}) -> {self.output_annotation}"

    async def __call__(self, data: _In) -> _Out:
        transform_exception = None

        transformed: _Out | None = None
        try:
            transformed = await self.transform_async(data)
        except Exception as exception:
            if type(exception.__cause__) == TransformerException:
                transform_exception = exception.__cause__
            else:
                tb = traceback.extract_tb(exception.__traceback__)

                # TODO: Make this filter condition stronger
                transformer_frames = [
                    frame
                    for frame in tb
                    if frame.name == self.__class__.__name__ or frame.name == "transform"
                ]

                if len(transformer_frames) == 1:
                    transformer_frame = transformer_frames[0]
                    exception_message = (
                        f"\n  "
                        f'File "{transformer_frame.filename}", line {transformer_frame.lineno}, '
                        f'in transformer "{self.__class__.__name__}"\n  '
                        f"  >> {transformer_frame.line}"
                    )
                else:
                    exception_message = (
                        f'An error occurred in transformer "{self.__class__.__name__}"'
                    )

                transform_exception = TransformerException(
                    internal_exception=exception,
                    raiser_transformer=self,
                    message=exception_message,
                )

        if transform_exception is not None:
            raise transform_exception.internal_exception

        if type(transformed) is not None:
            return cast(_Out, transformed)

        raise NotImplementedError

    def copy(
        self,
        transform: Callable[[BaseTransformer, _In], Awaitable[_Out]] | None = None,
        regenerate_instance_id: bool = False,
    ) -> "AsyncTransformer[_In, _Out]":
        copied = copy.copy(self)

        func_type = types.MethodType
        if transform is not None:
            setattr(copied, "transform_async", func_type(transform, copied))

        if regenerate_instance_id:
            copied.instance_id = uuid.uuid4()

        if self.previous is not None:
            # copy_next_previous = 'none' if copy_previous == 'first_previous' else copy_previous
            if type(self.previous) == tuple:
                new_previous: list[BaseTransformer] = [
                    previous_transformer.copy() for previous_transformer in self.previous
                ]
                copied._previous = cast(PreviousTransformer, tuple(new_previous))
            elif isinstance(self.previous, BaseTransformer):
                copied._previous = self.previous.copy()

        copied._children = [
            child.copy(regenerate_instance_id=True) for child in self.children
        ]

        return copied

    @overload
    def __rshift__(
        self, next_node: BaseTransformer[_Out, _NextOut, Any]
    ) -> "AsyncTransformer[_In, _NextOut]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            BaseTransformer[_Out, _NextOut, Any], BaseTransformer[_Out, _Out2, Any]
        ],
    ) -> "AsyncTransformer[_In, tuple[_NextOut, _Out2]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            BaseTransformer[_Out, _NextOut, Any],
            BaseTransformer[_Out, _Out2, Any],
            BaseTransformer[_Out, _Out3, Any],
        ],
    ) -> "AsyncTransformer[_In, tuple[_NextOut, _Out2, _Out3]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            BaseTransformer[_Out, _NextOut, Any],
            BaseTransformer[_Out, _Out2, Any],
            BaseTransformer[_Out, _Out3, Any],
            BaseTransformer[_Out, _Out4, Any],
        ],
    ) -> "AsyncTransformer[_In, tuple[_NextOut, _Out2, _Out3, _Out4]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            BaseTransformer[_Out, _NextOut, Any],
            BaseTransformer[_Out, _Out2, Any],
            BaseTransformer[_Out, _Out3, Any],
            BaseTransformer[_Out, _Out4, Any],
            BaseTransformer[_Out, _Out5, Any],
        ],
    ) -> "AsyncTransformer[_In, tuple[_NextOut, _Out2, _Out3, _Out4, _Out5]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            BaseTransformer[_Out, _NextOut, Any],
            BaseTransformer[_Out, _Out2, Any],
            BaseTransformer[_Out, _Out3, Any],
            BaseTransformer[_Out, _Out4, Any],
            BaseTransformer[_Out, _Out5, Any],
            BaseTransformer[_Out, _Out6, Any],
        ],
    ) -> "AsyncTransformer[_In, tuple[_NextOut, _Out2, _Out3, _Out4, _Out5, _Out6]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            BaseTransformer[_Out, _NextOut, Any],
            BaseTransformer[_Out, _Out2, Any],
            BaseTransformer[_Out, _Out3, Any],
            BaseTransformer[_Out, _Out4, Any],
            BaseTransformer[_Out, _Out5, Any],
            BaseTransformer[_Out, _Out6, Any],
            BaseTransformer[_Out, _Out7, Any],
        ],
    ) -> (
        "AsyncTransformer[_In, tuple[_NextOut, _Out2, _Out3, _Out4, _Out5, _Out6, _Out7]]"
    ):
        pass

    def __rshift__(self, next_node):
        pass



import asyncio
import os
import unittest
from collections.abc import Sequence
from pathlib import Path
from typing import Iterable, TypeVar, Any

from typing_extensions import assert_type

from tests.lib.conditioners import if_not_zero, if_is_even
from tests.lib.ensurers import is_even, same_value, same_value_int, is_greater_than_10
from tests.lib.transformers import (
    square,
    square_root,
    plus1,
    minus1,
    to_string,
    logarithm,
    repeat,
    format_currency,
    tuple_concatenate,
)
from gloe import (
    Transformer,
    transformer,
    partial_transformer,
    partial_async_transformer,
    ensure,
    async_transformer,
    AsyncTransformer,
)
from gloe.utils import forward
from gloe.experimental import bridge
from gloe.collection import Map
from mypy import api


In = TypeVar("In")
Out = TypeVar("Out")


class TestTransformerTypes(unittest.TestCase):
    def _test_transformer_simple_typing(self):
        """
        Test the most simple transformer typing
        """

        graph = square
        assert_type(graph, Transformer[float, float])

    def _test_simple_flow_typing(self):
        """
        Test the most simple transformer typing
        """

        graph = square >> square_root

        assert_type(graph, Transformer[float, float])

    def _test_flow_with_mixed_types(self):
        """
        Test the most simple transformer typing
        """

        graph = square >> square_root >> to_string

        assert_type(graph, Transformer[float, str])

    def _test_divergent_flow_types(self):
        """
        Test the most simple transformer typing
        """

        graph2 = square >> square_root >> (to_string, square)
        assert_type(graph2, Transformer[float, tuple[str, float]])

        graph3 = square >> square_root >> (to_string, square, to_string)
        assert_type(graph3, Transformer[float, tuple[str, float, str]])

        graph4 = square >> square_root >> (to_string, square, to_string, square)
        assert_type(graph4, Transformer[float, tuple[str, float, str, float]])

        graph5 = (
            square >> square_root >> (to_string, square, to_string, square, to_string)
        )
        assert_type(graph5, Transformer[float, tuple[str, float, str, float, str]])

        graph6 = (
            square
            >> square_root
            >> (to_string, square, to_string, square, to_string, square)
        )
        assert_type(graph6, Transformer[float, tuple[str, float, str, float, str, float]])

        graph7 = (
            square
            >> square_root
            >> (to_string, square, to_string, square, to_string, square, to_string)
        )
        assert_type(
            graph7, Transformer[float, tuple[str, float, str, float, str, float, str]]
        )

    def _test_conditioned_flow_types(self):
        """
        Test the most simple transformer typing
        """

        conditioned_graph = square >> square_root >> if_not_zero.Then(plus1).Else(minus1)

        assert_type(conditioned_graph, Transformer[float, float])

        conditioned_graph2 = (
            square >> square_root >> if_not_zero.Then(to_string).Else(square)
        )

        assert_type(conditioned_graph2, Transformer[float, str | float])

    def _test_chained_condition_flow_types(self):
        """
        Test the most simple transformer typing
        """

        chained_conditions_graph = (
            if_is_even.Then(square).ElseIf(lambda x: x < 10).Then(to_string).ElseNone()
        )

        assert_type(chained_conditions_graph, Transformer[float, float | str | None])

    def _test_partial_transformer(self):
        """
        Test the curried transformer typing
        """

        log2 = logarithm(base=2)
        assert_type(log2, Transformer[float, float])

        repeater = repeat(n_times=2, linebreak=True)
        assert_type(repeater, Transformer[str, str])

    def _test_transformer_init(self):
        """
        Test the transformer initializer typing
        """

        formatter = format_currency(thousands_separator=",")

        assert_type(formatter, Transformer[float, str])

    def _test_transformer_map(self):
        """
        Test the transformer map collection operation
        """

        mapped_logarithm = forward[list[float]]() >> Map(
            format_currency(thousands_separator=",")
        )

        assert_type(mapped_logarithm, Transformer[list[float], list[str]])

    def _test_transformer_ensurer(self):
        @ensure(incoming=[is_even])
        @transformer
        def t1(n1: int) -> int:
            return n1 * n1

        @ensure(outcome=[is_even])
        @transformer
        def t2(n1: int) -> int:
            return n1 * n1

        @ensure(changes=[same_value_int])
        @transformer
        def t3(n1: int) -> int:
            return n1 * n1

        @ensure(incoming=[is_even], outcome=[is_even])
        @transformer
        def t4(n1: int) -> int:
            return n1 * n1

        @ensure(incoming=[is_even], outcome=[is_greater_than_10], changes=[same_value])
        @transformer
        def t5(n1: int) -> float:
            return n1 * n1

        @ensure(outcome=[is_greater_than_10], changes=[same_value])
        @transformer
        def t6(num: float) -> float:
            return num

        assert_type(t1, Transformer[int, int])
        assert_type(t2, Transformer[int, int])
        assert_type(t3, Transformer[int, int])
        assert_type(t4, Transformer[int, int])
        assert_type(t5, Transformer[int, float])
        assert_type(t6, Transformer[float, float])

    def _test_transformer_init_ensurer(self):
        @ensure(incoming=[is_even])
        @partial_transformer
        def ti1(n1: int, n2: int) -> int:
            return n1 * n2

        @ensure(outcome=[is_even])
        @partial_transformer
        def ti2(n1: int, n2: int) -> int:
            return n1 * n2

        @ensure(changes=[same_value_int])
        @partial_transformer
        def ti3(n1: int, n2: int) -> int:
            return n1 * n2

        @ensure(incoming=[is_even], outcome=[is_even])
        @partial_transformer
        def ti4(n1: int, n2: int) -> int:
            return n1 * n2

        @ensure(incoming=[is_even], changes=[same_value_int])
        @partial_transformer
        def ti5(n1: int, n2: int) -> int:
            return n1 * n2

        assert_type(ti1(2), Transformer[int, int])
        assert_type(ti2(2), Transformer[int, int])
        assert_type(ti3(2), Transformer[int, int])
        assert_type(ti4(2), Transformer[int, int])
        assert_type(ti5(2), Transformer[int, int])

    def _test_bridge(self):
        num_bridge = bridge[float]("num")

        graph = plus1 >> num_bridge.pick() >> minus1 >> num_bridge.drop()

        assert_type(graph, Transformer[float, tuple[float, float]])

    def _test_async_transformer(self):
        @async_transformer
        async def _square(num: int) -> float:
            return float(num * num)

        async_pipeline = _square >> to_string
        async_pipeline2 = forward[int]() >> _square >> to_string
        async_pipeline3 = forward[int]() >> (_square, _square >> to_string)
        async_pipeline4 = _square >> (to_string, forward[float]())
        async_pipeline5 = _square >> (to_string, forward[float]()) >> tuple_concatenate

        assert_type(_square, AsyncTransformer[int, float])
        assert_type(async_pipeline, AsyncTransformer[int, str])
        assert_type(async_pipeline2, AsyncTransformer[int, str])
        assert_type(async_pipeline3, AsyncTransformer[int, tuple[float, str]])
        assert_type(async_pipeline4, AsyncTransformer[int, tuple[str, float]])
        assert_type(async_pipeline5, AsyncTransformer[int, str])

    def _test_partial_async_transformer(self):
        @partial_async_transformer
        async def sleep_and_forward(data: dict[str, str], delay: int) -> dict[str, str]:
            await asyncio.sleep(delay)
            return data

        pipeline = sleep_and_forward(1) >> forward()

        assert_type(pipeline, AsyncTransformer[dict[str, str], dict[str, str]])

    def test_all(self):
        file_path = Path(os.path.abspath(__file__))
        config_path = (file_path.parent.parent / "mypy.ini").absolute()
        result = api.run(
            [
                __file__,
                "--follow-imports=silent",
                f"--config-file={config_path}",
            ]
        )
        self.assertRegex(result[0], "Success")


if __name__ == "__main__":
    unittest.main()



from pygments.style import Style
from pygments.token import (
    Keyword,
    Name,
    Comment,
    String,
    Error,
    Number,
    Operator,
    Generic,
    Token,
    Whitespace,
)


class GloeDarkStyle(Style):
    name = "gloe-dark"

    background_color = "#202020"
    highlight_color = "#404040"
    line_number_color = "#aaaaaa"

    styles = {
        Token: "#d0d0d0",
        Whitespace: "#666666",
        Comment: "italic #ababab",
        Comment.Preproc: "noitalic bold #ff3a3a",
        Comment.Special: "noitalic bold #e50808 bg:#520000",
        Keyword: "bold #45df9a",
        Keyword.Pseudo: "nobold",
        Operator.Word: "bold #45df9a",
        String: "#6ad7ca",
        String.Other: "#6ad7ca",
        Number: "#51b2fd",
        Name.Builtin: "#2fbccd",
        Name.Variable: "#40ffff",
        Name.Constant: "#40ffff",
        Name.Class: "underline #14c8ef",
        Name.Function: "#14c8ef",
        Name.Namespace: "underline #14c8ef",
        Name.Exception: "#bbbbbb",
        Name.Tag: "bold #45df9a",
        Name.Attribute: "#bbbbbb",
        Name.Decorator: "#6ad7ca",
        Generic.Heading: "bold #ffffff",
        Generic.Subheading: "underline #ffffff",
        Generic.Deleted: "#ff3a3a",
        Generic.Inserted: "#589819",
        Generic.Error: "#ff3a3a",
        Generic.Emph: "italic",
        Generic.Strong: "bold",
        Generic.EmphStrong: "bold italic",
        Generic.Prompt: "#aaaaaa",
        Generic.Output: "#cccccc",
        Generic.Traceback: "#ff3a3a",
        Error: "bg:#e3d2d2 #a61717",
    }


class GloeLightStyle(Style):
    name = "gloe-light"

    background_color = "#ffffff"
    highlight_color = "#404040"
    line_number_color = "#aaaaaa"

    styles = {
        Token: "#d0d0d0",
        Whitespace: "#666666",
        Comment: "italic #ababab",
        Comment.Preproc: "noitalic bold #ff3a3a",
        Comment.Special: "noitalic bold #e50808 bg:#520000",
        Keyword: "bold #45df9a",
        Keyword.Pseudo: "nobold",
        Operator.Word: "bold #45df9a",
        String: "#6ad7ca",
        String.Other: "#6ad7ca",
        Number: "#51b2fd",
        Name.Builtin: "#2fbccd",
        Name.Variable: "#40ffff",
        Name.Constant: "#40ffff",
        Name.Class: "underline #14c8ef",
        Name.Function: "#14c8ef",
        Name.Namespace: "underline #14c8ef",
        Name.Exception: "#bbbbbb",
        Name.Tag: "bold #45df9a",
        Name.Attribute: "#bbbbbb",
        Name.Decorator: "#6ad7ca",
        Generic.Heading: "bold #ffffff",
        Generic.Subheading: "underline #ffffff",
        Generic.Deleted: "#ff3a3a",
        Generic.Inserted: "#589819",
        Generic.Error: "#ff3a3a",
        Generic.Emph: "italic",
        Generic.Strong: "bold",
        Generic.EmphStrong: "bold italic",
        Generic.Prompt: "#aaaaaa",
        Generic.Output: "#cccccc",
        Generic.Traceback: "#ff3a3a",
        Error: "bg:#e3d2d2 #a61717",
    }



from functools import wraps
from types import GenericAlias
from typing import (
    TypeVar,
    get_origin,
    TypeAlias,
    TypedDict,
    Generic,
    Union,
    _GenericAlias,
    ParamSpec,
    Callable,
    Awaitable,
)  # type: ignore


def _format_tuple(
    tuple_annotation: tuple, generic_input_param, input_annotation
) -> str:
    formatted: list[str] = []
    for annotation in tuple_annotation:
        formatted.append(
            _format_return_annotation(annotation, generic_input_param, input_annotation)
        )
    return f"({', '.join(formatted)})"


def _format_union(
    tuple_annotation: tuple, generic_input_param, input_annotation
) -> str:
    formatted: list[str] = []
    for annotation in tuple_annotation:
        formatted.append(
            _format_return_annotation(annotation, generic_input_param, input_annotation)
        )
    return f"({' | '.join(formatted)})"


def _format_generic_alias(
    return_annotation: GenericAlias, generic_input_param, input_annotation
) -> str:
    alias_name = return_annotation.__name__
    formatted: list[str] = []
    for annotation in return_annotation.__args__:
        formatted.append(
            _format_return_annotation(annotation, generic_input_param, input_annotation)
        )
    return f"{alias_name}[{', '.join(formatted)}]"


def _format_return_annotation(
    return_annotation, generic_input_param, input_annotation
) -> str:
    if type(return_annotation) == str:
        return return_annotation
    if type(return_annotation) == tuple:
        return _format_tuple(return_annotation, generic_input_param, input_annotation)
    if return_annotation.__name__ in {"tuple", "Tuple"}:
        return _format_tuple(
            return_annotation.__args__, generic_input_param, input_annotation
        )
    if return_annotation.__name__ in {"Union"}:
        return _format_union(
            return_annotation.__args__, generic_input_param, input_annotation
        )
    if (
        type(return_annotation) == GenericAlias
        or type(return_annotation) == _GenericAlias
    ):
        return _format_generic_alias(
            return_annotation, generic_input_param, input_annotation
        )

    if return_annotation == generic_input_param:
        return str(input_annotation.__name__)

    return str(return_annotation.__name__)


def _match_types(generic, specific, ignore_mismatches=True):
    if type(generic) == TypeVar:
        return {generic: specific}

    specific_origin = get_origin(specific)
    generic_origin = get_origin(generic)

    if specific_origin is None and generic_origin is None:
        return {}

    if (specific_origin is None or generic_origin is None) or not issubclass(
        specific_origin, generic_origin
    ):
        if ignore_mismatches:
            return {}
        raise Exception(f"Type {generic} does not match with {specific}")

    generic_args = getattr(generic, "__args__", None)
    specific_args = getattr(specific, "__args__", None)

    if specific_args is None and specific_args is None:
        return {}

    if generic_args is None:
        if ignore_mismatches:
            return {}
        raise Exception(f"Type {generic} in generic has no arguments")

    if specific_args is None:
        if ignore_mismatches:
            return {}
        raise Exception(f"Type {specific} in specific has no arguments")

    if len(generic_args) != len(specific_args):
        if ignore_mismatches:
            return {}
        raise Exception(
            f"Number of arguments of type {generic} is different in specific type"
        )

    matches = {}
    for generic_arg, specific_arg in zip(generic_args, specific_args):
        matched_types = _match_types(generic_arg, specific_arg)
        matches.update(matched_types)

    return matches


def _specify_types(generic, spec):
    if type(generic) == TypeVar:
        tp = spec.get(generic)
        if tp is None:
            return generic
        return tp

    generic_args = getattr(generic, "__args__", None)

    if generic_args is None:
        return generic

    origin = get_origin(generic)

    args = tuple(_specify_types(arg, spec) for arg in generic_args)

    return GenericAlias(origin, args)


_Args = ParamSpec("_Args")
_R = TypeVar("_R")


def awaitify(sync_func: Callable[_Args, _R]) -> Callable[_Args, Awaitable[_R]]:
    async def async_func(*args, **kwargs):
        return sync_func(*args, **kwargs)

    return async_func



import copy
import types
import uuid
import inspect
from functools import cached_property
from inspect import Signature

import networkx as nx
from networkx import DiGraph, Graph
from typing import (
    Any,
    Callable,
    Generic,
    TypeVar,
    Union,
    cast,
    Iterable,
    get_args,
    get_origin,
    TypeAlias,
    Type,
)
from uuid import UUID
from itertools import groupby

from gloe._utils import _format_return_annotation

__all__ = ["BaseTransformer", "TransformerException", "PreviousTransformer"]

_In = TypeVar("_In")
_Out = TypeVar("_Out")
_NextOut = TypeVar("_NextOut")
_Self = TypeVar("_Self", bound="BaseTransformer")

_Out2 = TypeVar("_Out2")
_Out3 = TypeVar("_Out3")
_Out4 = TypeVar("_Out4")
_Out5 = TypeVar("_Out5")
_Out6 = TypeVar("_Out6")
_Out7 = TypeVar("_Out7")


PreviousTransformer: TypeAlias = Union[
    None,
    _Self,
    tuple[_Self, _Self],
    tuple[_Self, _Self, _Self],
    tuple[_Self, _Self, _Self, _Self],
    tuple[_Self, _Self, _Self, _Self, _Self],
    tuple[_Self, _Self, _Self, _Self, _Self, _Self],
    tuple[_Self, _Self, _Self, _Self, _Self, _Self, _Self],
]


class TransformerException(Exception):
    def __init__(
        self,
        internal_exception: Union["TransformerException", Exception],
        raiser_transformer: "BaseTransformer",
        message: str | None = None,
    ):
        self._internal_exception = internal_exception
        self.raiser_transformer = raiser_transformer
        self._traceback = internal_exception.__traceback__
        internal_exception.__cause__ = self
        super().__init__(message)

    @property
    def internal_exception(self):
        return self._internal_exception.with_traceback(self._traceback)


class BaseTransformer(Generic[_In, _Out, _Self]):
    def __init__(self):
        self._previous: PreviousTransformer["BaseTransformer"] = None
        self._children: list["BaseTransformer"] = []
        self._invisible = False
        self.id = uuid.uuid4()
        self.instance_id = uuid.uuid4()
        self._label = self.__class__.__name__
        self._graph_node_props: dict[str, Any] = {"shape": "box"}
        self.events = []

    @property
    def label(self) -> str:
        """
        Label used in visualization.

        When the transformer is created by the `@transformer` decorator, it is the
        name of the function.

        When creating a transformer by extending the `Transformer` class, it is the name of
        the class.
        """
        return self._label

    @property
    def graph_node_props(self) -> dict[str, Any]:
        return self._graph_node_props

    @property
    def children(self) -> list["BaseTransformer"]:
        return self._children

    @property
    def previous(self) -> PreviousTransformer["BaseTransformer"]:
        """
        Previous transformers. It can be None, a single transformer, or a tuple of many
        transformers.
        """
        return self._previous

    @property
    def invisible(self) -> bool:
        return self._invisible

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other):
        if isinstance(other, BaseTransformer):
            return self.id == other.id
        return NotImplemented

    def copy(
        self,
        transform: Callable[[_Self, _In], _Out] | None = None,
        regenerate_instance_id: bool = False,
    ) -> _Self:
        copied = copy.copy(self)

        func_type = types.MethodType
        if transform is not None:
            setattr(copied, "transform", func_type(transform, copied))

        if regenerate_instance_id:
            copied.instance_id = uuid.uuid4()

        if self.previous is not None:
            # copy_next_previous = 'none' if copy_previous == 'first_previous' else copy_previous
            if type(self.previous) == tuple:
                new_previous: list[BaseTransformer] = [
                    previous_transformer.copy() for previous_transformer in self.previous
                ]
                copied._previous = cast(PreviousTransformer, tuple(new_previous))
            elif isinstance(self.previous, BaseTransformer):
                copied._previous = self.previous.copy()

        copied._children = [
            child.copy(regenerate_instance_id=True) for child in self.children
        ]

        return copied

    @property
    def graph_nodes(self) -> dict[UUID, "BaseTransformer"]:
        nodes = {self.instance_id: self}

        if self.previous is not None:
            if type(self.previous) == tuple:
                for prev in self.previous:
                    nodes = {**nodes, **prev.graph_nodes}
            elif isinstance(self.previous, BaseTransformer):
                nodes = {**nodes, **self.previous.graph_nodes}

        for child in self.children:
            nodes = {**nodes, **child.graph_nodes}

        return nodes

    def _set_previous(self, previous: PreviousTransformer):
        if self.previous is None:
            self._previous = previous
        elif type(self.previous) == tuple:
            for previous_transformer in self.previous:
                previous_transformer._set_previous(previous)
        elif isinstance(self.previous, BaseTransformer):
            self.previous._set_previous(previous)

    def signature(self) -> Signature:
        return self._signature(type(self))

    def _signature(self, klass: Type) -> Signature:
        orig_bases = getattr(self, "__orig_bases__", [])
        transformer_args = [
            get_args(base) for base in orig_bases if get_origin(base) == klass
        ]
        generic_args = [
            get_args(base) for base in orig_bases if get_origin(base) == Generic
        ]

        orig_class = getattr(self, "__orig_class__", None)

        specific_args = {}
        if (
            len(transformer_args) == 1
            and len(generic_args) == 1
            and orig_class is not None
        ):
            generic_arg = generic_args[0]
            transformer_arg = transformer_args[0]
            specific_args = {
                generic: specific
                for generic, specific in zip(generic_arg, get_args(orig_class))
                if generic in transformer_arg
            }

        signature = inspect.signature(self.transform)
        new_return_annotation = specific_args.get(
            signature.return_annotation, signature.return_annotation
        )
        parameters = list(signature.parameters.values())
        if len(parameters) > 0:
            parameter = parameters[0]
            parameter = parameter.replace(
                annotation=specific_args.get(parameter.annotation, parameter.annotation)
            )
            return signature.replace(
                return_annotation=new_return_annotation,
                parameters=[parameter],
            )

        return signature.replace(return_annotation=new_return_annotation)

    @property
    def output_type(self) -> Any:
        signature = self.signature()
        return signature.return_annotation

    @property
    def output_annotation(self) -> str:
        output_type = self.output_type

        return_type = _format_return_annotation(output_type, None, None)
        return return_type

    @property
    def input_type(self) -> Any:
        parameters = list(self.signature().parameters.items())
        if len(parameters) > 0:
            parameter_type = parameters[0][1].annotation
            return parameter_type

    @property
    def input_annotation(self) -> str:
        return self.input_type.__name__

    def _add_net_node(self, net: Graph, custom_data: dict[str, Any] = {}):
        node_id = self.node_id
        props = {**self.graph_node_props, **custom_data, "label": self.label}
        if node_id not in net.nodes:
            net.add_node(node_id, **props)
        else:
            nx.set_node_attributes(net, {node_id: props})
        return node_id

    def _add_child_node(
        self,
        child: "BaseTransformer",
        child_net: DiGraph,
        parent_id: str,
        next_node: "BaseTransformer",
    ):
        child._dag(child_net, next_node, custom_data={"parent_id": parent_id})

    @property
    def node_id(self) -> str:
        return str(self.instance_id)

    @cached_property
    def visible_previous(self) -> PreviousTransformer["BaseTransformer"]:
        previous = self.previous

        if isinstance(previous, BaseTransformer):
            if previous.invisible:
                if previous.previous is None:
                    return previous

                if type(previous.previous) == tuple:
                    return previous.previous

                return previous.visible_previous
            else:
                return previous

        return previous

    def _add_children_subgraph(self, net: DiGraph, next_node: "BaseTransformer"):
        next_node_id = next_node.node_id
        children_nets = [DiGraph() for _ in self.children]
        visible_previous = self.visible_previous

        for child, child_net in zip(self.children, children_nets):
            self._add_child_node(child, child_net, self.node_id, next_node)
            net.add_nodes_from(child_net.nodes.data())
            net.add_edges_from(child_net.edges.data())

            child_root_node = [n for n in child_net.nodes if child_net.in_degree(n) == 0][
                0
            ]
            child_final_node = [
                n for n in child_net.nodes if child_net.out_degree(n) == 0
            ][0]

            if self.invisible:
                if type(visible_previous) == tuple:
                    for prev in visible_previous:
                        net.add_edge(
                            prev.node_id, child_root_node, label=prev.output_annotation
                        )
                elif isinstance(visible_previous, BaseTransformer):
                    net.add_edge(
                        visible_previous.node_id,
                        child_root_node,
                        label=visible_previous.output_annotation,
                    )
            else:
                node_id = self._add_net_node(net)
                net.add_edge(node_id, child_root_node)

            if child_final_node != next_node_id:
                net.add_edge(
                    child_final_node, next_node_id, label=next_node.input_annotation
                )

    def _dag(
        self,
        net: DiGraph,
        next_node: Union["BaseTransformer", None] = None,
        custom_data: dict[str, Any] = {},
    ):
        in_nodes = [edge[1] for edge in net.in_edges()]

        previous = self.previous
        if previous is not None:
            if type(previous) == tuple:
                if self.invisible and next_node is not None:
                    next_node_id = next_node._add_net_node(net)
                    _next_node = next_node
                else:
                    next_node_id = self._add_net_node(net, custom_data)
                    _next_node = self

                for prev in previous:
                    previous_node_id = prev.node_id

                    # TODO: check the impact of the below line to the Mapper transformer
                    if not prev.invisible and len(prev.children) == 0:
                        net.add_edge(
                            previous_node_id, next_node_id, label=prev.output_annotation
                        )

                    if previous_node_id not in in_nodes:
                        prev._dag(net, _next_node, custom_data)
            elif isinstance(previous, BaseTransformer):
                if self.invisible and next_node is not None:
                    next_node_id = next_node._add_net_node(net)
                    _next_node = next_node
                else:
                    next_node_id = self._add_net_node(net, custom_data)
                    _next_node = self

                previous_node_id = previous.node_id

                if len(previous.children) == 0 and (
                    not previous.invisible or previous.previous is None
                ):
                    previous_node_id = previous._add_net_node(net)
                    net.add_edge(
                        previous_node_id, next_node_id, label=previous.output_annotation
                    )

                if previous_node_id not in in_nodes:
                    previous._dag(net, _next_node, custom_data)
        else:
            self._add_net_node(net, custom_data)

        if len(self.children) > 0 and next_node is not None:
            self._add_children_subgraph(net, next_node)

    def graph(self) -> DiGraph:
        net = nx.DiGraph()
        net.graph["splines"] = "ortho"
        self._dag(net)
        return net

    # pragma: not covered
    def export(self, path: str, with_edge_labels: bool = True):
        net = self.graph()
        boxed_nodes = [
            node
            for node in net.nodes.data()
            if "parent_id" in node[1] and "bounding_box" in node[1]
        ]
        if not with_edge_labels:
            for u, v in net.edges:
                net.edges[u, v]["label"] = ""

        agraph = nx.nx_agraph.to_agraph(net)
        subgraphs: Iterable[tuple] = groupby(boxed_nodes, key=lambda x: x[1]["parent_id"])
        for parent_id, nodes in subgraphs:
            nodes = list(nodes)
            node_ids = [node[0] for node in nodes]
            if len(nodes) > 0:
                label = nodes[0][1]["box_label"]
                agraph.add_subgraph(
                    node_ids, label=label, name=f"cluster_{parent_id}", style="dotted"
                )
        agraph.write(path)

    def __len__(self):
        return 1



from typing import Generic, Iterable, TypeVar

from networkx import DiGraph

from ..transformers import Transformer

_T = TypeVar("_T")
_S = TypeVar("_S")
_U = TypeVar("_U")


class MapOver(Generic[_T, _U], Transformer[_T, list[_U]]):
    def __init__(
        self,
        iterable: list[_S],
        mapping_transformer: Transformer[tuple[_T, _S], _U],
    ):
        super().__init__()
        self.iterable = iterable
        self.mapping_transformer = mapping_transformer
        self._invisible = True
        self._children = [mapping_transformer]

    def transform(self, data: _T) -> list[_U]:
        lopping_result = []
        for item in self.iterable:
            lopping_result.append(self.mapping_transformer((data, item)))
        return lopping_result

    def _add_child_node(
        self,
        child: "Transformer",
        child_net: DiGraph,
        parent_id: str,
        next_node: "Transformer",
    ):
        child._dag(
            child_net,
            next_node,
            {"parent_id": parent_id, "bounding_box": True, "box_label": "mapping"},
        )



import unittest

from tests.lib.conditioners import if_is_even, if_not_zero
from tests.lib.transformers import square, square_root, plus1, minus1, to_string


class TestConditionerTransformer(unittest.TestCase):
    def test_conditioner_transformer_name(self):
        """
        Test the name of a ConditionerTransformer
        """
        graph = if_not_zero.Then(plus1).Else(minus1)
        self.assertEqual("if_not_zero", graph.__class__.__name__)

    def test_conditioned_flow(self):
        """
        Test the most simple conditioned case
        """

        conditioned_graph = (
            square >> square_root >> if_not_zero.Then(plus1).Else(minus1)
        )

        self.assertEqual(conditioned_graph(1), 2)
        self.assertEqual(conditioned_graph(0), -1)

    def test_else_none(self):
        """
        Test the else None case
        """
        graph = if_not_zero.Then(plus1).ElseNone()
        self.assertEqual(2, graph(1))
        self.assertIsNone(graph(0))

    def test_flow_with_many_conditions(self):
        """
        Test the case of a flow with many conditions
        """

        conditioned_graph = (
            square
            >> square_root
            >> (
                if_not_zero.Then(plus1 >> minus1).Else(minus1),
                if_not_zero.Then(to_string).Else(to_string),
            )
        )

        self.assertEqual(conditioned_graph(1), (1.0, "1.0"))
        self.assertEqual(conditioned_graph(0), (-1.0, "0.0"))

    def test_chained_conditions(self):
        """
        Test the case of a flow with chained conditions
        """

        conditioned_graph = square >> (
            if_is_even.Then(plus1 >> minus1)
            .ElseIf(lambda x: x == 1.0)
            .Then(plus1 >> plus1)
            .ElseIf(lambda x: x == 25.0)
            .Then(minus1 >> minus1)
            .Else(square_root)
        )

        self.assertEqual(3.0, conditioned_graph(1.0))
        self.assertEqual(3.0, conditioned_graph(3.0))
        self.assertEqual(16.0, conditioned_graph(4.0))
        self.assertEqual(23.0, conditioned_graph(5.0))

    def test_conditioner_transformer_length(self):
        """
        Test length property of conditioner transformers
        """

        conditioned_graph = square >> (
            if_is_even.Then(plus1 >> minus1)
            .ElseIf(lambda x: x == 1.0)
            .Then(plus1 >> plus1)
            .ElseIf(lambda x: x == 25.0)
            .Then(minus1 >> minus1)
            .Else(square_root)
        )

        self.assertEqual(8.0, len(conditioned_graph))


if __name__ == "__main__":
    unittest.main()



import unittest

from gloe.collection import Map, MapOver
from tests.lib.transformers import square, plus1, sum_tuple2


class TestTransformerCollections(unittest.TestCase):
    def test_transformer_map(self):
        """
        Test the mapping transformer
        """

        mapping1 = Map(square) >> Map(plus1)
        mapping2 = Map(square >> plus1)

        seq = [10.0, 9.0, 3.0, 2.0, -1.0]
        result1 = list(mapping1(seq))
        result2 = list(mapping2(seq))

        expected = [101.0, 82.0, 10.0, 5.0, 2.0]
        self.assertListEqual(expected, result1)
        self.assertListEqual(expected, result2)

    def test_transformer_map_over(self):
        """
        Test the mapping transformer
        """

        data = [10.0, 9.0, 3.0, 2.0, -1.0]
        mapping = MapOver(data, sum_tuple2) >> Map(plus1)

        result = list(mapping(-1.0))

        self.assertListEqual(result, data)



from typing import Any, Tuple, TypeVar, Generic

from gloe.functional import transformer
from gloe.transformers import Transformer

__all__ = ["forget", "debug", "forward", "forward_incoming"]

_In = TypeVar("_In")
_Out = TypeVar("_Out")


@transformer
def forget(data: Any) -> None:
    """Transform any input data to `None`"""
    return None


@transformer
def debug(incoming: _In) -> _In:
    breakpoint()
    return incoming


class forward(Generic[_In], Transformer[_In, _In]):
    def __init__(self):
        super().__init__()
        self._invisible = True

    def __repr__(self):
        if self.previous is not None:
            return str(self.previous)

        return super().__repr__()

    def transform(self, data: _In) -> _In:
        return data


def forward_incoming(
    inner_transformer: Transformer[_In, _Out]
) -> Transformer[_In, Tuple[_Out, _In]]:
    return forward[_In]() >> (inner_transformer, forward())



import unittest

from gloe.utils import forward, forward_incoming
from tests.lib.transformers import sum_tuple2


class TestTransformerUtils(unittest.TestCase):
    def test_forward_incoming(self):
        test_forward = forward_incoming(sum_tuple2)

        self.assertEqual(test_forward((3, 3)), (6, (3, 3)))

    def test_forward_repr(self):
        test_forward = forward[int]()

        self.assertEqual("int -> (forward) -> int", repr(test_forward))



from dataclasses import dataclass
from inspect import Signature
from types import GenericAlias, UnionType
from typing import Callable, Generic, Optional, TypeVar, Union

from gloe.transformers import Transformer
from gloe.utils import forget


In = TypeVar("In")
ThenOut = TypeVar("ThenOut")
ElseOut = TypeVar("ElseOut")
ElseIfOut = TypeVar("ElseIfOut")
PrevThenOut = TypeVar("PrevThenOut")


@dataclass
class _Implication(Generic[In, ThenOut]):
    condition: Callable[[In], bool]
    then_transformer: Transformer[In, ThenOut]


class ConditionerTransformer(
    Generic[In, ThenOut, ElseOut], Transformer[In, Union[ThenOut, ElseOut]]
):
    def __init__(
        self,
        implications: list[_Implication[In, ThenOut]],
        else_transformer: Transformer[In, ElseOut],
    ):
        super().__init__()
        self.implications = implications
        self.else_transformer = else_transformer
        self._graph_node_props = {"shape": "diamond", "style": "filled", "port": "n"}
        self._children = [
            *[impl.then_transformer for impl in implications],
            else_transformer,
        ]

    def transform(self, data: In) -> Union[ThenOut, ElseOut]:
        for implication in self.implications:
            if implication.condition(data):
                return implication.then_transformer.transform(data)

        return self.else_transformer.transform(data)

    def signature(self) -> Signature:
        else_signature: Signature = self.else_transformer.signature()
        return_signature: list[Signature] = [
            impl.then_transformer.signature().return_annotation
            for impl in self.implications
        ] + [else_signature.return_annotation]
        new_signature = else_signature.replace(
            return_annotation=GenericAlias(UnionType, tuple(return_signature))
        )
        return new_signature

    def copy(
        self,
        transform: Callable[["Transformer", In], Union[ThenOut, ElseOut]] | None = None,
        regenerate_instance_id: bool = False,
    ) -> "ConditionerTransformer[In, ThenOut, ElseOut]":
        copied: ConditionerTransformer[In, ThenOut, ElseOut] = super().copy(
            transform, regenerate_instance_id
        )
        copied.implications = [
            _Implication(
                impl.condition, impl.then_transformer.copy(regenerate_instance_id=True)
            )
            for impl in copied.implications
        ]
        copied.else_transformer = self.else_transformer.copy(regenerate_instance_id=True)
        copied._children = [
            *[impl.then_transformer for impl in copied.implications],
            copied.else_transformer,
        ]
        return copied

    def __len__(self):
        then_transformers_len = [len(impl.then_transformer) for impl in self.implications]
        return sum(then_transformers_len) + len(self.else_transformer)


class _IfThen(Generic[In, ThenOut, PrevThenOut]):
    def __init__(
        self,
        implication: _Implication[In, ThenOut],
        name: str,
        prev_implications: list[_Implication[In, PrevThenOut]] = [],
    ):
        super().__init__()
        self._implication = implication
        self._prev_implications = prev_implications
        self._name = name
        self._implications: list[
            _Implication[In, Union[ThenOut, PrevThenOut]]
        ] = self._prev_implications + [self._implication]

    def Else(
        self, else_transformer: Transformer[In, ElseOut]
    ) -> Transformer[In, Union[ThenOut, PrevThenOut, ElseOut]]:
        new_transformer: ConditionerTransformer[
            In, Union[ThenOut, PrevThenOut], ElseOut
        ] = ConditionerTransformer(self._implications, else_transformer)

        new_transformer.__class__.__name__ = self._name
        new_transformer._label = self._name

        return new_transformer

    def ElseIf(
        self, condition: Callable[[In], bool]
    ) -> "_ElseIf[In, Union[ThenOut, PrevThenOut]]":
        else_if: "_ElseIf[In, Union[ThenOut, PrevThenOut]]" = _ElseIf(
            condition, self._implications, self._name
        )
        else_if.__class__.__name__ = self.__class__.__name__
        return else_if

    def ElseNone(
        self,
    ) -> Transformer[In, Optional[Union[ThenOut, PrevThenOut]]]:
        new_transformer: ConditionerTransformer[
            In, Union[ThenOut, PrevThenOut], None
        ] = ConditionerTransformer(self._implications, forget)
        new_transformer.__class__.__name__ = self.__class__.__name__
        return new_transformer


class _ElseIf(Generic[In, PrevThenOut]):
    def __init__(
        self,
        condition: Callable[[In], bool],
        prev_implications: list[_Implication[In, PrevThenOut]],
        name: str,
    ):
        super().__init__()
        self._condition = condition
        self._prev_implications: list[_Implication[In, PrevThenOut]] = prev_implications
        self._name = name

    def Then(
        self, next_transformer: Transformer[In, ElseIfOut]
    ) -> _IfThen[In, ElseIfOut, PrevThenOut]:
        implication = _Implication[In, ElseIfOut](self._condition, next_transformer)
        if_then = _IfThen[In, ElseIfOut, PrevThenOut](
            implication, self._name, self._prev_implications
        )
        return if_then


class If(Generic[In]):
    """
    It is used to start a condition chaining

    Example:
        The below example send a different email for admin and non admin users::

            send_email = (
                If(lambda user: "admin" in user.roles)
                    .Then(fetch_admin_data >> send_admin_email)
                .Else(send_member_email)
            )

    Args:
        condition: callable returning a boolean.
        name: optional argument that adds a label to condition node during plotting.
    """

    def __init__(self, condition: Callable[[In], bool], name: str | None = None):
        super().__init__()
        self._condition = condition
        self._name: str = name or condition.__name__

    def Then(
        self, next_transformer: Transformer[In, ThenOut]
    ) -> _IfThen[In, ThenOut, ThenOut]:
        """ """
        implication = _Implication[In, ThenOut](self._condition, next_transformer)
        if_then = _IfThen[In, ThenOut, ThenOut](implication, name=self._name)
        return if_then


def condition(func: Callable[[In], bool]) -> If[In]:
    """
    The condition decorator is responsible to build an instance of :code:`If[T]`
    class given a :code:`Callable[[T], bool]`.

    When instantiating the :code:`If` class, the parameter :code:`name` will be the name
    of the callable.

    See Also:
        For additional information about conditions and examples of its usage, consult
        :ref:`conditional-flows`.

    Example:
        The below example send a different email for admin and non admin users::

            @condition
            def is_admin(user: User) -> bool:
                return "admin" in user.roles

            send_email = (
                is_admin.Then(fetch_admin_data >> send_admin_email)
                .Else(send_member_email)
            )

    Args:
        func: A callable that check somenting about the incoming data and returns a
            boolean.
    """
    condition = If(func, func.__name__)
    return condition



from typing import Generic, TypeVar

from gloe.transformers import Transformer

_T = TypeVar("_T")
_S = TypeVar("_S")
_U = TypeVar("_U")


class Map(Generic[_T, _U], Transformer[list[_T], list[_U]]):
    """
    Transformer used to map values in a list using other transformers instead of
    functions.

    Unfortunately, we must use :code:`list` instead of a more general type like
    :code:`Iterable`, because neither Python nor Mypy recognizes :code:`list[T]` as a
    subclass of :code:`Iterable[T]`. A consequence of this limitation is::

        get_user_posts: Transformer[User, list[Post]]
        filter_archived_posts: Transformer[Iterable[Post], Iterable[Post]]

        get_user_posts >> filter_archived_posts
        # ERROR: Unsupported operand types for >> ("Transformer[User, list[Post]]" and "Transformer[Iterable[Post], Iterable[Post]]")

    To solve this problem, we need to use `Higher Kinded Types
    <https://sobolevn.me/2020/10/higher-kinded-types-in-python>`_, but this is only currently possible using
    `Mypy Plugins <https://mypy.readthedocs.io/en/stable/extending_mypy.html>`_.
    Gloe aims to avoid the need for such solutions.

    Example:
        In this example, we fetch a list of users that belongs to a group.
        Then, we get the posts of each user and finally turn it into a single list
        of posts.::

            @transformer
            def get_user_posts(user: User) -> list[Post]: ...

            get_posts_by_group: Transformer[Group, list[Post]] = (
                get_users_by_group >> Map(get_user_posts) >> flatten
            )
    Args:
        mapping_transformer: transformer applied to each item of the
            input list the yield the mapped item of the output list.
    """

    def __init__(self, mapping_transformer: Transformer[_T, _U]):
        super().__init__()
        self.mapping_transformer = mapping_transformer
        self._invisible = True
        self._children = [mapping_transformer]

        mapping_transformer._graph_node_props = mapping_transformer.graph_node_props | {
            "parent_id": self.instance_id,
            "bounding_box": True,
            "box_label": "mapping",
        }

    def transform(self, data: list[_T]) -> list[_U]:
        """
        Args:
            data: incoming list to be mapped. The items of this list must be of
                type :code:`_T`.

        Returns:
            The mapped list. The items of this new list are of type :code:`_U`.
        """
        mapping_result = []
        for item in data:
            mapping_result.append(self.mapping_transformer(item))
        return mapping_result



from contextvars import ContextVar
from typing import Generic, TypeVar

from gloe.transformers import Transformer

T = TypeVar("T")
B = TypeVar("B")


class EmptyBridgeOnDrop(Exception):
    def __init__(self, bridge_name: str):
        super().__init__(
            f"The bridge {bridge_name} has tried to be dropped but its value was not initialized"
        )


class _pick(Generic[T], Transformer[T, T]):
    def __init__(self, variable: ContextVar[T]):
        super().__init__()
        self._invisible = True
        self.variable = variable

    def transform(self, data: T) -> T:
        self.variable.set(data)
        return data


class _drop(Generic[B, T], Transformer[B, tuple[B, T]]):
    def __init__(self, variable: ContextVar[T]):
        super().__init__()
        self._invisible = True
        self._variable = variable

    def transform(self, data: B) -> tuple[B, T]:
        current_value: T
        try:
            current_value = self._variable.get()
        except LookupError:
            raise EmptyBridgeOnDrop(self._variable.name)
        return data, current_value


class bridge(Generic[T]):
    def __init__(self, name: str):
        self._variable: ContextVar[T] = ContextVar(name)

    def pick(self) -> Transformer[T, T]:
        return _pick(self._variable)

    def drop(self) -> Transformer[B, tuple[B, T]]:
        drop: _drop[B, T] = _drop(self._variable)
        return drop



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import sys

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
sys.path.insert(0, os.path.abspath("../.."))
sys.path.insert(0, os.path.abspath("pygments"))

project = "Gloe"
copyright = "2023, Samir Braga"
author = "Samir Braga"
release = "0.4.3"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx_toolbox.more_autodoc.variables",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    # "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinxext.opengraph",
    # "sphinx_autodoc_typehints",
    "myst_parser",
    "sphinx_copybutton",
]
overloads_location = "bottom"
napoleon_google_docstring = True
autosectionlabel_prefix_document = True
napoleon_use_rtype = False
# intersphinx_mapping = {"httpx": ("https://www.python-httpx.org/", None)}
ogp_social_cards = {
    "enable": True,
    "line_color": "#00afac",
    "image": "_static/assets/gloe-logo.png",
}
ogp_site_url = "https://gloe.ideos.com.br/"
# ogp_image = "https://gloe.ideos.com.br/_static/assets/gloe-logo.png"

templates_path = ["_templates"]
exclude_patterns = ["Thumbs.db", ".DS_Store"]
autodoc_typehints = "description"
autodoc_type_aliases = {
    "PreviousTransformer": "gloe.base_transformer.PreviousTransformer"
}
# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_title = "Gloe"
# html_logo = "assets/gloe-logo-small.png"
html_theme = "furo"
html_last_updated_fmt = ""
# html_use_index = False  # Don't create index
# html_domain_indices = False  # Don't need module indices
# html_copy_source = False  # Don't need sources
html_sidebars: dict[str, list[str]] = {
    "Home": ["/"],
}
# autodoc_default_options = {"ignore-module-all": True}

html_static_path = ["_static"]
html_css_files = ["theme_customs.css"]
html_favicon = "_static/assets/favicon.ico"
html_theme_options = {
    # "main_nav_links": {"Docs": "/index", "About": "/about"},
    "light_logo": "assets/gloe-logo-small.png",
    "dark_logo": "assets/gloe-logo-small.png",
    "dark_css_variables": {
        "color-brand-primary": "#00e6bf",
        "color-brand-content": "#00e6bf",
        "font-stack": "Roboto, sans-serif",
        "font-stack--monospace": "Courier, monospace",
        "font-size--normal": "Courier, monospace",
    },
    "light_css_variables": {
        "color-brand-primary": "#00afac",
        "color-brand-content": "#00afac",
        "font-stack": "Roboto, sans-serif",
        "font-stack--monospace": "Courier, monospace",
        "font-size--normal": "Courier, monospace",
    },
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/ideos/gloe",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}

# pygments_style = "styles.GloeStyle"
pygments_dark_style = "styles.GloeDarkStyle"
pygments_light_style = "styles.GloeLightStyle"



import asyncio
import unittest
from typing import cast

from tests.lib.transformers import (
    square,
    square_root,
    sum_tuple2,
    sum_tuple3,
    plus1,
    minus1,
    natural_logarithm,
    logarithm,
    LnOfNegativeNumber,
)
from gloe import (
    TransformerException,
    UnsupportedTransformerArgException,
    transformer,
)


class TestFunctionTransformer(unittest.TestCase):
    def test_transformer_wrong_signature(self):
        with self.assertWarns(RuntimeWarning):

            @transformer  # type: ignore
            def many_args(arg1: str, arg2: int):
                return arg1, arg2

    def test_transformer_hash(self):
        self.assertEqual(hash(square.id), square.__hash__())

    def test_linear_flow(self):
        """
        Test the most simple linear case
        """

        linear_graph = square >> square_root

        integer = 10
        result = linear_graph(integer)

        self.assertEqual(integer, result)

    def test_unsupported_argument(self):
        def just_a_normal_function():
            return None

        with self.assertRaises(
            UnsupportedTransformerArgException,
            msg=f"Unsupported transformer argument: {just_a_normal_function}",
        ):
            _ = square >> just_a_normal_function  # type: ignore

        with self.assertRaises(
            UnsupportedTransformerArgException,
            msg=f"Unsupported transformer argument: {just_a_normal_function}",
        ):
            _ = square >> (just_a_normal_function, plus1)  # type: ignore

    def test_previous_property(self):
        """
        Test the previous property
        """

        linear_graph = square >> square_root

        self.assertEqual(linear_graph.previous, square)

        divergent_graph = (
            square
            >> square_root
            >> (square >> (square_root, square_root), square >> square_root >> square)
        )

        self.assertIsNone(
            divergent_graph.previous[0].previous[0].previous.previous.previous.previous
        )

    def test_divergence_flow(self):
        """
        Test the most simple divergent case
        """

        divergent_graph = square >> (square_root, square)

        integer = 10
        result = divergent_graph(integer)

        self.assertEqual(result, (integer, 10000))

    def test_convergent_flow(self):
        """
        Test the most simple convergent case
        """

        convergent_graph = square >> (square_root, square) >> sum_tuple2

        integer = 10
        result = convergent_graph(integer)

        self.assertEqual(10000 + 10, result)

    def test_divergent_many_branches_flow(self):
        """
        Test the divergent case with many branches
        """

        convergent_graph = square >> (
            square_root,
            square,
            square_root,
            square,
            square_root,
            square,
        )

        integer = 10
        result = convergent_graph(integer)

        self.assertEqual(result, (10, 10000, 10, 10000, 10, 10000))

    def test_dynamically_created_flow(self):
        """
        Test the dynamically created graph using a iteration
        """

        graph = square >> square_root

        for i in range(10):
            graph = graph >> square >> square_root

        result = graph(10)
        self.assertEqual(result, 10)

    def test_exhausting_large_flow(self):
        """
        Test the instantiation of large graph
        """
        graph = plus1

        max_iters = 475
        for i in range(max_iters):
            graph = graph >> plus1

        result = graph(0)
        self.assertEqual(result, max_iters + 1)

    def test_recursive_flow(self):
        """
        Test the instantiation of large graph
        """
        graph = plus1 >> plus1
        graph = graph >> graph
        graph = graph >> graph
        graph = graph >> graph
        graph = graph >> graph
        graph = graph >> graph
        graph = graph >> graph
        graph = graph >> graph

        result = graph(0)
        self.assertEqual(result, 256)

    def test_graph_length_property(self):
        graph = square >> square_root

        for i in range(160):
            graph = graph >> square >> square_root

        self.assertEqual(len(graph), 160 * 2 + 2)

        graph2 = (
            square
            >> square_root
            >> (square >> square_root, square >> square_root, square >> square_root)
            >> sum_tuple3
            >> square
            >> square_root
            >> (square >> square_root, square >> square_root)
        )

        self.assertEqual(len(graph2), 15)

    def test_transformer_equality(self):
        graph = square >> square_root
        self.assertEqual(square, square)
        self.assertEqual(square, square.copy())
        self.assertNotEqual(graph, square_root)
        self.assertNotEqual(square, square_root)

    def test_transformer_pydoc_keeping(self):
        @transformer
        def to_string(num: int) -> str:
            """
            This transformer receives a number as input and return its representation as a string
            """
            return str(num)

        self.assertEqual(
            to_string.__doc__,
            """
            This transformer receives a number as input and return its representation as a string
            """,
        )

    def test_transformer_signature_representation(self):
        signature = square.signature()

        self.assertEqual(str(signature), "(num: float) -> float")

    # def test_transformer_nodes_retrieve(self):
    #     graph = square >> square_root >> square >> square_root
    #     nodes = graph.graph_nodes()
    #
    #     expected_nodes = {
    #         square.id: square,
    #         square_root.id: square_root
    #     }
    #     self.assertDictEqual(expected_nodes, nodes)
    #
    #     graph2 = square >> square_root >> (
    #         square >> square_root,
    #         square >> square_root
    #     )
    #     nodes = graph2.graph_nodes()
    #
    #     expected_nodes = {
    #         square.id: square,
    #         square_root.id: square_root
    #     }
    #     self.assertDictEqual(expected_nodes, nodes)

    def test_transformer_error_forward(self):
        """
        Test if an error raised inside a transformer can be caught outside it
        """
        graph = minus1 >> natural_logarithm
        self.assertRaises(LnOfNegativeNumber, lambda: graph(0))

    def test_transformer_error_handling(self):
        """
        Test if a raised error stores the correct TransformerException as its cause
        """

        graph = minus1 >> natural_logarithm
        try:
            graph(0)
        except LnOfNegativeNumber as exception:
            self.assertEqual(type(exception.__cause__), TransformerException)

            exception_ctx = cast(TransformerException, exception.__cause__)
            self.assertEqual(natural_logarithm, exception_ctx.raiser_transformer)

    def test_partial_transformer(self):
        """
        Test the curried transformer
        """

        graph = logarithm(base=2)
        self.assertEqual(graph(2), 1)
        self.assertEqual(graph.label, "logarithm")

    def test_transformers_on_a_running_event_loop(self):
        async def run_main():
            graph = square >> square_root
            graph(9)

        loop = asyncio.new_event_loop()
        loop.run_until_complete(run_main())


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Any

from networkx import DiGraph

from gloe import Transformer
from gloe.utils import forward

from tests.lib.conditioners import if_is_even
from tests.lib.transformers import (
    square,
    square_root,
    plus1,
    minus1,
    divide_by_2,
    times2,
    sum_tuple2,
    sum_tuple3,
    mul_tuple2,
    natural_logarithm,
    logarithm,
    to_string,
    tuple_concatenate,
)


class TestTransformerGraph(unittest.TestCase):
    def _get_nodes_by_name(self, transformer: Transformer) -> dict[str, str]:
        ids_by_name = {
            node.__class__.__name__: str(id)
            for id, node in transformer.graph_nodes.items()
        }

        return ids_by_name

    def _assert_graph_has_edges(
        self,
        transformer: Transformer,
        graph: DiGraph,
        expected_edges: list[tuple[str, str]],
    ):
        ids_by_name = self._get_nodes_by_name(transformer)

        edges = [edge for edge, props in list(graph.edges.items())]

        edges_with_id = [
            (ids_by_name[edge[0]], ids_by_name[edge[1]]) for edge in expected_edges
        ]

        for edge in edges_with_id:
            self.assertIn(edge, edges)

    def _assert_edge_properties(
        self,
        transformer: Transformer,
        graph: DiGraph,
        expected_edges_props: dict[tuple[str, str], Any],
    ):
        ids_by_name = self._get_nodes_by_name(transformer)
        edges_with_id = {
            (ids_by_name[edge[0]], ids_by_name[edge[1]]): props
            for edge, props in expected_edges_props.items()
        }
        current_edge_props = {(u, v): graph.get_edge_data(u, v) for u, v in edges_with_id}

        for edge, current_props in current_edge_props.items():
            expected_props = edges_with_id[edge]
            self.assertDictEqual(current_props | expected_props, current_props)

    def _assert_nodes_count(self, expected: int, graph: DiGraph):
        nodes = graph.nodes.items()
        self.assertEqual(expected, len(nodes))

    def _assert_edges_count(self, expected: int, graph: DiGraph):
        edges = [edge for edge, props in list(graph.edges.items())]
        self.assertEqual(expected, len(edges))

    def _assert_nodes_properties(
        self,
        nodes_properties: dict[str, dict[str, Any]],
        transformer: Transformer,
        graph: DiGraph,
    ):
        ids_by_name = self._get_nodes_by_name(transformer)
        props_by_id = dict(graph.nodes)
        for name, properties in nodes_properties.items():
            node_id = ids_by_name[name]
            current_props = props_by_id[node_id]
            expected_props = nodes_properties[name]
            self.assertDictEqual(current_props | expected_props, current_props)

    def test_simplest_case(self):
        identity = square >> square_root >> plus1 >> minus1
        graph: DiGraph = identity.graph()

        self._assert_nodes_count(4, graph)
        self._assert_edges_count(3, graph)

        expected_edges = [
            ("square", "square_root"),
            ("square_root", "plus1"),
            ("plus1", "minus1"),
        ]

        self._assert_graph_has_edges(identity, graph, expected_edges)

    def test_simple_divergent_case(self):
        divergent = square >> (plus1, minus1) >> sum_tuple2

        graph: DiGraph = divergent.graph()

        self._assert_nodes_count(5, graph)

        # Five edges:
        #           +---->  plus1  ----+
        #  square --+                  +--> (Converge) --> sum_tuple2
        #           +----> minus1 -----+
        self._assert_edges_count(5, graph)

        expected_edges = [
            ("square", "plus1"),
            ("square", "minus1"),
            ("plus1", "Converge"),
            ("minus1", "Converge"),
            ("Converge", "sum_tuple2"),
        ]

        self._assert_graph_has_edges(divergent, graph, expected_edges)

    def test_complex_divergent_case(self):
        divergent = (
            square
            >> (
                plus1 >> if_is_even.Then(square_root).Else(forward()),
                minus1 >> natural_logarithm,
                times2 >> divide_by_2,
            )
            >> sum_tuple3
        )

        graph: DiGraph = divergent.graph()

        self._assert_nodes_count(11, graph)  # Each divergent connection has a hidden node
        self._assert_edges_count(13, graph)

        expected_edges = [
            ("square", "plus1"),
            ("square", "minus1"),
            ("square", "times2"),
            ("plus1", "if_is_even"),
            ("if_is_even", "square_root"),
            ("if_is_even", "forward"),
            ("square_root", "Converge"),
            ("forward", "Converge"),
            ("minus1", "natural_logarithm"),
            ("times2", "divide_by_2"),
            ("natural_logarithm", "Converge"),
            ("divide_by_2", "Converge"),
            ("Converge", "sum_tuple3"),
        ]

        self._assert_graph_has_edges(divergent, graph, expected_edges)

    def test_nested_divergent_case(self):
        divergent = (
            square
            >> (
                square_root >> plus1 >> (minus1, natural_logarithm),
                times2 >> divide_by_2,
            )
            >> sum_tuple2
        )

        graph: DiGraph = divergent.graph()

        self._assert_nodes_count(10, graph)  # Each divergent connection has a hidden node
        self._assert_edges_count(11, graph)

        expected_edges = [
            ("square", "square_root"),
            ("square", "times2"),
            ("square_root", "plus1"),
            ("plus1", "minus1"),
            ("plus1", "natural_logarithm"),
            ("times2", "divide_by_2"),
        ]

        self._assert_graph_has_edges(divergent, graph, expected_edges)

    def test_simple_conditional_case(self):
        conditional = square_root >> if_is_even.Then(plus1).Else(forward()) >> minus1

        graph: DiGraph = conditional.graph()

        self._assert_nodes_count(5, graph)
        self._assert_edges_count(5, graph)

        expected_edges = [
            ("square_root", "if_is_even"),
            ("if_is_even", "plus1"),
            ("if_is_even", "forward"),
            ("plus1", "minus1"),
            ("forward", "minus1"),
        ]

        self._assert_graph_has_edges(conditional, graph, expected_edges)

    def test_complex_conditional_case(self):
        then_graph = plus1 >> square >> (times2, divide_by_2) >> sum_tuple2
        conditional = square_root >> if_is_even.Then(then_graph).Else(forward()) >> minus1

        graph: DiGraph = conditional.graph()

        self._assert_nodes_count(10, graph)
        self._assert_edges_count(11, graph)

        expected_edges = [
            ("square_root", "if_is_even"),
            ("if_is_even", "plus1"),
            ("if_is_even", "forward"),
            ("plus1", "square"),
            ("square", "times2"),
            ("square", "divide_by_2"),
            ("times2", "Converge"),
            ("divide_by_2", "Converge"),
            ("Converge", "sum_tuple2"),
            ("sum_tuple2", "minus1"),
            ("forward", "minus1"),
        ]

        self._assert_graph_has_edges(conditional, graph, expected_edges)

    def test_repeated_nodes_case(self):
        repeated = (
            plus1 >> plus1 >> (plus1 >> plus1 >> plus1, minus1 >> minus1) >> sum_tuple2
        )

        graph: DiGraph = repeated.graph()

        self._assert_nodes_count(9, graph)
        self._assert_edges_count(9, graph)

    def test_begin_node_case(self):
        begin1 = forward[float]() >> (plus1, minus1) >> sum_tuple2

        graph: DiGraph = begin1.graph()

        self._assert_nodes_count(5, graph)
        self._assert_edges_count(5, graph)

        expected_edges = [
            ("forward", "plus1"),
            ("forward", "minus1"),
            ("plus1", "Converge"),
            ("minus1", "Converge"),
            ("Converge", "sum_tuple2"),
        ]

        self._assert_graph_has_edges(begin1, graph, expected_edges)

    def test_invisible_nodes_case(self):
        begin1 = forward[float]() >> (plus1, minus1) >> sum_tuple2

        begin2 = forward[float]() >> (divide_by_2, times2) >> mul_tuple2

        begin3 = begin1 >> begin2

        graph: DiGraph = begin3.graph()

        self._assert_nodes_count(9, graph)
        self._assert_edges_count(10, graph)

    def test_transformer_init_case(self):
        init_graph = logarithm(2) >> square

        graph: DiGraph = init_graph.graph()

        self._assert_nodes_count(2, graph)
        self._assert_edges_count(1, graph)

        expected_edges = [
            ("logarithm", "square"),
        ]

        self._assert_graph_has_edges(init_graph, graph, expected_edges)

    def test_nodes_properties_case(self):
        then_graph = plus1 >> square >> (times2, divide_by_2) >> sum_tuple2
        conditional = square_root >> if_is_even.Then(then_graph).Else(forward()) >> minus1

        graph = conditional.graph()

        box_nodes = [
            "plus1",
            "square",
            "times2",
            "divide_by_2",
            "sum_tuple2",
            "square_root",
            "minus1",
        ]
        box_nodes_properties = {
            node: {"label": node, "shape": "box"} for node in box_nodes
        }
        nodes_properties = {
            **box_nodes_properties,
            "Converge": {"label": "", "shape": "diamond"},
            "if_is_even": {"label": "if_is_even", "shape": "diamond"},
        }

        self._assert_nodes_properties(nodes_properties, conditional, graph)

    def test_edge_labels_case(self):
        single_edge = plus1 >> square
        graph = single_edge.graph()

        expected_edge_props = {("plus1", "square"): {"label": "float"}}

        self._assert_edge_properties(single_edge, graph, expected_edge_props)

        divergent_edges = plus1 >> (square >> to_string, square_root) >> tuple_concatenate
        graph = divergent_edges.graph()

        expected_edge_props = {
            ("plus1", "square"): {"label": "float"},
            ("plus1", "square_root"): {"label": "float"},
            ("square", "to_string"): {"label": "float"},
            ("to_string", "Converge"): {"label": "str"},
            ("square_root", "Converge"): {"label": "float"},
            ("Converge", "tuple_concatenate"): {"label": "(str, float)"},
        }

        self._assert_edge_properties(divergent_edges, graph, expected_edge_props)



import asyncio
import unittest
from typing import TypeVar
from gloe import async_transformer, ensure
from gloe.functional import partial_async_transformer
from gloe.utils import forward

_In = TypeVar("_In")

_DATA = {"foo": "bar"}


@async_transformer
async def request_data(url: str) -> dict[str, str]:
    await asyncio.sleep(0.1)
    return _DATA


class HasNotBarKey(Exception):
    pass


def has_bar_key(dict: dict[str, str]):
    if "bar" not in dict.keys():
        raise HasNotBarKey()


_URL = "http://my-service"


class TestAsyncTransformer(unittest.IsolatedAsyncioTestCase):
    async def test_basic_case(self):
        test_forward = request_data >> forward()

        result = await test_forward(_URL)

        self.assertDictEqual(result, _DATA)

    async def test_begin_with_transformer(self):
        test_forward = forward[str]() >> request_data

        result = await test_forward(_URL)

        self.assertDictEqual(result, _DATA)

    async def test_async_on_divergent_connection(self):
        test_forward = forward[str]() >> (forward[str](), request_data)

        result = await test_forward(_URL)

        self.assertEqual(result, (_URL, _DATA))

    async def test_divergent_connection_from_async(self):
        test_forward = request_data >> (
            forward[dict[str, str]](),
            forward[dict[str, str]](),
        )

        result = await test_forward(_URL)

        self.assertEqual(result, (_DATA, _DATA))

    async def test_partial_async_transformer(self):
        @partial_async_transformer
        async def sleep_and_forward(
            data: dict[str, str], delay: float
        ) -> dict[str, str]:
            await asyncio.sleep(delay)
            return data

        pipeline = sleep_and_forward(0.1) >> forward()

        result = await pipeline(_DATA)

        self.assertEqual(result, _DATA)

    async def test_ensure_async_transformer(self):
        @ensure(outcome=[has_bar_key])
        @async_transformer
        async def ensured_request(url: str) -> dict[str, str]:
            await asyncio.sleep(0.1)
            return _DATA

        pipeline = ensured_request >> forward()

        with self.assertRaises(HasNotBarKey):
            await pipeline(_URL)

    async def test_ensure_partial_async_transformer(self):
        @ensure(outcome=[has_bar_key])
        @partial_async_transformer
        async def ensured_delayed_request(url: str, delay: float) -> dict[str, str]:
            await asyncio.sleep(delay)
            return _DATA

        pipeline = ensured_delayed_request(0.1) >> forward()

        with self.assertRaises(HasNotBarKey):
            await pipeline(_URL)



class NumberIsOdd(Exception):
    pass


class NumberIsEven(Exception):
    pass


class NumberLessThanOrEquals10(Exception):
    pass


class NumbersNotEqual(Exception):
    pass


class LnOfNegativeNumber(Exception):
    def __init__(self, num: float):
        super().__init__(f"The natural logarithm of the negative number {num} is not a real number")



import traceback
from abc import ABC, abstractmethod
from inspect import Signature

from typing import (
    TypeVar,
    overload,
    cast,
    Any,
    TypeAlias,
    Union,
)

from gloe.base_transformer import BaseTransformer, TransformerException
from gloe.async_transformer import AsyncTransformer

__all__ = ["Transformer"]

I = TypeVar("I")
O = TypeVar("O")
O1 = TypeVar("O1")
O2 = TypeVar("O2")
O3 = TypeVar("O3")
O4 = TypeVar("O4")
O5 = TypeVar("O5")
O6 = TypeVar("O6")
O7 = TypeVar("O7")


Tr: TypeAlias = "Transformer"
AT: TypeAlias = AsyncTransformer
BT: TypeAlias = BaseTransformer[I, O, Any]

AsyncNext2 = Union[
    tuple[AT[O, O1], BT[O, O2]],
    tuple[BT[O, O1], AT[O, O2]],
]

AsyncNext3 = Union[
    tuple[AT[O, O1], BT[O, O2], BT[O, O3]],
    tuple[BT[O, O1], AT[O, O2], BT[O, O3]],
    tuple[BT[O, O1], BT[O, O2], AT[O, O3]],
]

AsyncNext4 = Union[
    tuple[AT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4]],
    tuple[BT[O, O1], AT[O, O2], BT[O, O3], BT[O, O4]],
    tuple[BT[O, O1], BT[O, O2], AT[O, O3], BT[O, O4]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], AT[O, O4]],
]

AsyncNext5 = Union[
    tuple[AT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5]],
    tuple[BT[O, O1], AT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5]],
    tuple[BT[O, O1], BT[O, O2], AT[O, O3], BT[O, O4], BT[O, O5]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], AT[O, O4], BT[O, O5]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], AT[O, O5]],
]

AsyncNext6 = Union[
    tuple[AT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5], BT[O, O6]],
    tuple[BT[O, O1], AT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5], BT[O, O6]],
    tuple[BT[O, O1], BT[O, O2], AT[O, O3], BT[O, O4], BT[O, O5], BT[O, O6]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], AT[O, O4], BT[O, O5], BT[O, O6]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], AT[O, O5], BT[O, O6]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5], AT[O, O6]],
]

AsyncNext7 = Union[
    tuple[AT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5], BT[O, O6], BT[O, O7]],
    tuple[BT[O, O1], AT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5], BT[O, O6], BT[O, O7]],
    tuple[BT[O, O1], BT[O, O2], AT[O, O3], BT[O, O4], BT[O, O5], BT[O, O6], BT[O, O7]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], AT[O, O4], BT[O, O5], BT[O, O6], BT[O, O7]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], AT[O, O5], BT[O, O6], BT[O, O7]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5], AT[O, O6], BT[O, O7]],
    tuple[BT[O, O1], BT[O, O2], BT[O, O3], BT[O, O4], BT[O, O5], BT[O, O6], AT[O, O7]],
]


class Transformer(BaseTransformer[I, O, "Transformer"], ABC):
    """
    A Transformer is the generic block with the responsibility to take an input of type
    `T` and transform it to an output of type `S`.

    See Also:
        Read more about this feature in the page :ref:`creating-a-transformer`.

    Example:
        Typical usage example::

            class Stringifier(Transformer[dict, str]):
                ...

    """

    def __init__(self):
        super().__init__()
        self.__class__.__annotations__ = self.transform.__annotations__

    @abstractmethod
    def transform(self, data: I) -> O:
        """Main method to be implemented and responsible to perform the transformer logic"""

    def signature(self) -> Signature:
        return self._signature(Transformer)

    def __repr__(self):
        return f"{self.input_annotation} -> ({type(self).__name__}) -> {self.output_annotation}"

    def __call__(self, data: I) -> O:
        transform_exception = None

        transformed: O | None = None
        try:
            transformed = self.transform(data)
        except Exception as exception:
            if type(exception.__cause__) == TransformerException:
                transform_exception = exception.__cause__
            else:
                tb = traceback.extract_tb(exception.__traceback__)

                # TODO: Make this filter condition stronger
                transformer_frames = [
                    frame
                    for frame in tb
                    if frame.name == self.__class__.__name__ or frame.name == "transform"
                ]

                if len(transformer_frames) == 1:
                    transformer_frame = transformer_frames[0]
                    exception_message = (
                        f"\n  "
                        f'File "{transformer_frame.filename}", line {transformer_frame.lineno}, '
                        f'in transformer "{self.__class__.__name__}"\n  '
                        f"  >> {transformer_frame.line}"
                    )
                else:
                    exception_message = (
                        f'An error occurred in transformer "{self.__class__.__name__}"'
                    )

                transform_exception = TransformerException(
                    internal_exception=exception,
                    raiser_transformer=self,
                    message=exception_message,
                )

        if transform_exception is not None:
            raise transform_exception.internal_exception

        if type(transformed) is not None:
            return cast(O, transformed)

        raise NotImplementedError

    @overload
    def __rshift__(
        self,
        next_node: tuple["Tr[O, O1]", "Tr[O, O2]"],
    ) -> "Tr[I, tuple[O1, O2]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple["Tr[O, O1]", "Tr[O, O2]", "Tr[O, O3]"],
    ) -> "Transformer[I, tuple[O1, O2, O3]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple["Tr[O, O1]", "Tr[O, O2]", "Tr[O, O3]", "Tr[O, O4]"],
    ) -> "Tr[I, tuple[O1, O2, O3, O4]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple["Tr[O, O1]", "Tr[O, O2]", "Tr[O, O3]", "Tr[O, O4]", "Tr[O, O5]"],
    ) -> "Tr[I, tuple[O1, O2, O3, O4, O5]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            "Tr[O, O1]", "Tr[O, O2]", "Tr[O, O3]", "Tr[O, O4]", "Tr[O, O5]", "Tr[O, O6]"
        ],
    ) -> "Tr[I, tuple[O1, O2, O3, O4, O5, O6]]":
        pass

    @overload
    def __rshift__(
        self,
        next_node: tuple[
            "Tr[O, O1]",
            "Tr[O, O2]",
            "Tr[O, O3]",
            "Tr[O, O4]",
            "Tr[O, O5]",
            "Tr[O, O6]",
            "Tr[O, O7]",
        ],
    ) -> "Tr[I, tuple[O1, O2, O3, O4, O5, O6, O7]]":
        pass

    @overload
    def __rshift__(self, next_node: "Tr[O, O1]") -> "Tr[I, O1]":
        pass

    @overload
    def __rshift__(self, next_node: AsyncTransformer[O, O1]) -> AsyncTransformer[I, O1]:
        pass

    @overload
    def __rshift__(
        self,
        next_node: AsyncNext2[O, O1, O2],
    ) -> AsyncTransformer[I, tuple[O1, O2]]:
        pass

    @overload
    def __rshift__(
        self,
        next_node: AsyncNext3[O, O1, O2, O3],
    ) -> AsyncTransformer[I, tuple[O1, O2, O3]]:
        pass

    @overload
    def __rshift__(
        self,
        next_node: AsyncNext4[O, O1, O2, O3, O4],
    ) -> AsyncTransformer[I, tuple[O1, O2, O3, O4]]:
        pass

    @overload
    def __rshift__(
        self,
        next_node: AsyncNext5[O, O1, O2, O3, O4, O5],
    ) -> AsyncTransformer[I, tuple[O1, O2, O3, O4, O5]]:
        pass

    @overload
    def __rshift__(
        self,
        next_node: AsyncNext6[O, O1, O2, O3, O4, O5, O6],
    ) -> AsyncTransformer[I, tuple[O1, O2, O3, O4, O5, O6]]:
        pass

    @overload
    def __rshift__(
        self,
        next_node: AsyncNext7[O, O1, O2, O3, O4, O5, O6, O7],
    ) -> AsyncTransformer[I, tuple[O1, O2, O3, O4, O5, O6, O7]]:
        pass

    def __rshift__(self, next_node):
        pass



from gloe._composition_utils import _compose_nodes
from gloe.functional import (
    transformer,
    partial_transformer,
    partial_async_transformer,
    async_transformer,
)
from gloe.conditional import If, condition
from gloe.ensurer import ensure
from gloe.exceptions import UnsupportedTransformerArgException
from gloe.transformers import Transformer
from gloe.base_transformer import BaseTransformer, PreviousTransformer
from gloe.base_transformer import TransformerException
from gloe.async_transformer import AsyncTransformer

__all__ = [
    "transformer",
    "partial_transformer",
    "partial_async_transformer",
    "async_transformer",
    "If",
    "condition",
    "ensure",
    "UnsupportedTransformerArgException",
    "BaseTransformer",
    "PreviousTransformer",
    "Transformer",
    "TransformerException",
    "AsyncTransformer",
]

setattr(Transformer, "__rshift__", _compose_nodes)
setattr(AsyncTransformer, "__rshift__", _compose_nodes)



import inspect
from abc import abstractmethod, ABC
from types import FunctionType
from typing import Any, Callable, Generic, ParamSpec, Sequence, TypeVar, cast, overload

from gloe.async_transformer import AsyncTransformer
from gloe.functional import _PartialTransformer, _PartialAsyncTransformer
from gloe.transformers import Transformer

_T = TypeVar("_T")
_S = TypeVar("_S")
_U = TypeVar("_U")
_P1 = ParamSpec("_P1")


class TransformerEnsurer(Generic[_T, _S], ABC):
    @abstractmethod
    def validate_input(self, data: _T):
        """Perform a validation on incoming data before execute the transformer code"""

    @abstractmethod
    def validate_output(self, data: _T, output: _S):
        """Perform a validation on outcome data after execute the transformer code"""

    def __call__(self, transformer: Transformer[_T, _S]) -> Transformer[_T, _S]:
        def transform(this: Transformer, data: _T) -> _S:
            self.validate_input(data)
            output = transformer.transform(data)
            self.validate_output(data, output)
            return output

        transformer_cp = transformer.copy(transform)
        return transformer_cp


def input_ensurer(func: Callable[[_T], Any]) -> TransformerEnsurer[_T, Any]:
    class LambdaEnsurer(TransformerEnsurer[_T, _S]):
        __doc__ = func.__doc__
        __annotations__ = cast(FunctionType, func).__annotations__

        def validate_input(self, data: _T):
            func(data)

        def validate_output(self, data: _T, output: _S):
            pass

    return LambdaEnsurer()


@overload
def output_ensurer(func: Callable[[_T, _S], Any]) -> TransformerEnsurer[_T, _S]:
    pass


@overload
def output_ensurer(func: Callable[[_S], Any]) -> TransformerEnsurer[Any, _S]:
    pass


def output_ensurer(func: Callable):
    class LambdaEnsurer(TransformerEnsurer):
        __doc__ = func.__doc__
        __annotations__ = cast(FunctionType, func).__annotations__

        def validate_input(self, data):
            pass

        def validate_output(self, data, output):
            if len(inspect.signature(func).parameters) == 1:
                func(output)
            else:
                func(data, output)

    return LambdaEnsurer()


class _ensure_base:
    @overload
    def __call__(self, transformer: Transformer[_U, _S]) -> Transformer[_U, _S]:
        pass

    @overload
    def __call__(
        self, transformer_init: _PartialTransformer[_T, _P1, _U]
    ) -> _PartialTransformer[_T, _P1, _U]:
        pass

    @overload
    def __call__(self, transformer: AsyncTransformer[_U, _S]) -> AsyncTransformer[_U, _S]:
        pass

    @overload
    def __call__(
        self, transformer_init: _PartialAsyncTransformer[_T, _P1, _U]
    ) -> _PartialAsyncTransformer[_T, _P1, _U]:
        pass

    def __call__(self, arg):
        if isinstance(arg, Transformer):
            return self._generate_new_transformer(arg)
        if isinstance(arg, AsyncTransformer):
            return self._generate_new_async_transformer(arg)
        if isinstance(arg, _PartialTransformer):
            transformer_init = arg

            def ensured_transformer_init(*args, **kwargs):
                transformer = transformer_init(*args, **kwargs)
                return self._generate_new_transformer(transformer)

            return ensured_transformer_init
        if isinstance(arg, _PartialAsyncTransformer):
            async_transformer_init = arg

            def ensured_async_transformer_init(*args, **kwargs):
                async_transformer = async_transformer_init(*args, **kwargs)
                return self._generate_new_async_transformer(async_transformer)

            return ensured_async_transformer_init

    @abstractmethod
    def _generate_new_transformer(self, transformer: Transformer) -> Transformer:
        pass

    @abstractmethod
    def _generate_new_async_transformer(
        self, transformer: AsyncTransformer
    ) -> AsyncTransformer:
        pass


class _ensure_incoming(Generic[_T], _ensure_base):
    def __init__(self, incoming: Sequence[Callable[[_T], Any]]):
        self.input_ensurers_instances = [input_ensurer(ensurer) for ensurer in incoming]

    def _generate_new_transformer(self, transformer: Transformer) -> Transformer:
        def transform(_, data):
            for ensurer in self.input_ensurers_instances:
                ensurer.validate_input(data)
            output = transformer.transform(data)
            return output

        transformer_cp = transformer.copy(transform)
        return transformer_cp

    def _generate_new_async_transformer(
        self, transformer: AsyncTransformer
    ) -> AsyncTransformer:
        async def transform_async(_, data):
            for ensurer in self.input_ensurers_instances:
                ensurer.validate_input(data)
            output = await transformer.transform_async(data)
            return output

        transformer_cp = transformer.copy(transform_async)

        return transformer_cp


class _ensure_outcome(Generic[_S], _ensure_base):
    def __init__(self, incoming: Sequence[Callable[[_S], Any]]):
        self.output_ensurers_instances = [output_ensurer(ensurer) for ensurer in incoming]

    def _generate_new_transformer(self, transformer: Transformer) -> Transformer:
        def transform(_, data):
            output = transformer.transform(data)
            for ensurer in self.output_ensurers_instances:
                ensurer.validate_output(data, output)
            return output

        transformer_cp = transformer.copy(transform)

        return transformer_cp

    def _generate_new_async_transformer(
        self, transformer: AsyncTransformer
    ) -> AsyncTransformer:
        async def transform_async(_, data):
            output = await transformer.transform_async(data)
            for ensurer in self.output_ensurers_instances:
                ensurer.validate_output(data, output)
            return output

        transformer_cp = transformer.copy(transform_async)

        return transformer_cp


class _ensure_changes(Generic[_T, _S], _ensure_base):
    def __init__(self, changes: Sequence[Callable[[_T, _S], Any]]):
        self.changes_ensurers_instances = [output_ensurer(ensurer) for ensurer in changes]

    def _generate_new_transformer(self, transformer: Transformer) -> Transformer:
        def transform(_, data):
            output = transformer.transform(data)
            for ensurer in self.changes_ensurers_instances:
                ensurer.validate_output(data, output)
            return output

        transformer_cp = transformer.copy(transform)

        return transformer_cp

    def _generate_new_async_transformer(
        self, transformer: AsyncTransformer
    ) -> AsyncTransformer:
        async def transform_async(_, data):
            output = await transformer.transform_async(data)
            for ensurer in self.changes_ensurers_instances:
                ensurer.validate_output(data, output)
            return output

        transformer_cp = transformer.copy(transform_async)

        return transformer_cp


class _ensure_both(Generic[_T, _S], _ensure_base):
    def __init__(
        self,
        incoming: Sequence[Callable[[_T], Any]],
        outcome: Sequence[Callable[[_S], Any]],
        changes: Sequence[Callable[[_T, _S], Any]],
    ):
        incoming_seq = incoming if type(incoming) == list else [incoming]
        self.input_ensurers_instances = [
            input_ensurer(ensurer) for ensurer in incoming_seq
        ]

        outcome_seq = outcome if type(outcome) == list else [outcome]
        self.output_ensurers_instances = [
            output_ensurer(ensurer) for ensurer in outcome_seq
        ]

        changes_seq = changes if type(changes) == list else [changes]
        self.output_ensurers_instances = self.output_ensurers_instances + [
            output_ensurer(ensurer) for ensurer in changes_seq
        ]

    def _generate_new_transformer(self, transformer: Transformer) -> Transformer:
        def transform(_, data):
            for ensurer in self.input_ensurers_instances:
                ensurer.validate_input(data)
            output = transformer.transform(data)
            for ensurer in self.output_ensurers_instances:
                ensurer.validate_output(data, output)
            return output

        transformer_cp = transformer.copy(transform)
        return transformer_cp

    def _generate_new_async_transformer(
        self, transformer: AsyncTransformer
    ) -> AsyncTransformer:
        async def transform_async(_, data):
            for ensurer in self.input_ensurers_instances:
                ensurer.validate_input(data)
            output = await transformer.transform_async(data)
            for ensurer in self.output_ensurers_instances:
                ensurer.validate_output(data, output)
            return output

        transformer_cp = transformer.copy(transform_async)
        return transformer_cp


@overload
def ensure(incoming: Sequence[Callable[[_T], Any]]) -> _ensure_incoming[_T]:
    pass


@overload
def ensure(outcome: Sequence[Callable[[_S], Any]]) -> _ensure_outcome[_S]:
    pass


@overload
def ensure(changes: Sequence[Callable[[_T, _S], Any]]) -> _ensure_changes[_T, _S]:
    pass


@overload
def ensure(
    incoming: Sequence[Callable[[_T], Any]], outcome: Sequence[Callable[[_S], Any]]
) -> _ensure_both[_T, _S]:
    pass


@overload
def ensure(
    incoming: Sequence[Callable[[_T], Any]], changes: Sequence[Callable[[_T, _S], Any]]
) -> _ensure_both[_T, _S]:
    pass


@overload
def ensure(
    outcome: Sequence[Callable[[_T], Any]], changes: Sequence[Callable[[_T, _S], Any]]
) -> _ensure_both[_T, _S]:
    pass


@overload
def ensure(
    incoming: Sequence[Callable[[_T], Any]],
    outcome: Sequence[Callable[[_S], Any]],
    changes: Sequence[Callable[[_T, _S], Any]],
) -> _ensure_both[_T, _S]:
    pass


def ensure(*args, **kwargs):
    """
    This decorator is used in transformers to ensure some validation based on its incoming
    data, outcome data, or both.

    These validations are performed by validators. Validators are simple callable
    functions that validate certain aspects of the input, output, or the differences
    between them. If the validation fails, it must raise an exception.

    The decorator :code:`@ensure` returns some intermediate classes to assist with the
    internal logic of Gloe. However, the result of applying it to a transformer is just
    a new transformer with the exact same attributes, but it includes an additional
    validation layer.

    The motivation of the many overloads is just to allow the user to define different types
    of validators interchangeably.

    See also:
        For more detailed information about this feature, refer to the :ref:`ensurers` page.

    Args:
        incoming (Sequence[Callable[[_T], Any]]): sequence of validators that will be
            applied to the incoming data. The type :code:`_T` refers to the incoming type.
            Defaut value: :code:`[]`.
        outcome (Sequence[Callable[[_S], Any]]): sequence of validators that will be
            applied to the outcome data. The type :code:`_S` refers to the outcome type.
            Defaut value: :code:`[]`.
        changes (Sequence[Callable[[_T, _S], Any]]): sequence of validators that will be
            applied to both incoming and outcome data. The type :code:`_T` refers to the
            incoming type, and type :code:`_S` refers to the outcome type.
            Defaut value: :code:`[]`.
    """
    if len(kwargs.keys()) == 1 and "incoming" in kwargs:
        return _ensure_incoming(kwargs["incoming"])

    if len(kwargs.keys()) == 1 and "outcome" in kwargs:
        return _ensure_outcome(kwargs["outcome"])

    if len(kwargs.keys()) == 1 and "changes" in kwargs:
        return _ensure_changes(kwargs["changes"])

    if len(kwargs.keys()) > 1:
        incoming = []
        if "incoming" in kwargs:
            incoming = kwargs["incoming"]

        outcome = []
        if "outcome" in kwargs:
            outcome = kwargs["outcome"]

        changes = []
        if "changes" in kwargs:
            changes = kwargs["changes"]

        return _ensure_both(incoming, outcome, changes)



import unittest

from gloe.experimental import EmptyBridgeOnDrop, bridge
from tests.lib.transformers import plus1, minus1


class TestTransformerBridges(unittest.TestCase):
    def test_basic_case(self):
        num_bridge = bridge[float]("num")

        graph = plus1 >> num_bridge.pick() >> minus1 >> num_bridge.drop()

        self.assertEqual((10.0, 11.0), graph(10.0))

    def test_error_case(self):
        num_bridge = bridge[float]("num")

        graph = plus1 >> minus1 >> num_bridge.drop()

        self.assertRaises(EmptyBridgeOnDrop, lambda: graph(0))



import math

from tests.lib.exceptions import LnOfNegativeNumber
from gloe import transformer, partial_transformer


@transformer
def square_root(num: float) -> float:
    return math.sqrt(num)


@transformer
def plus1(num: float) -> float:
    """
    Sum 1 to the number
    """
    return num + 1


@transformer
def square(num: float) -> float:
    return num * num


@transformer
def minus1(num: float) -> float:
    return num - 1


@transformer
def times2(num: float) -> float:
    return num * 2


@transformer
def divide_by_2(num: float) -> float:
    return num / 2


@transformer
def sum_tuple2(num: tuple[float, float]) -> float:
    return num[0] + num[1]


@transformer
def mul_tuple2(num: tuple[float, float]) -> float:
    return num[0] * num[1]


@transformer
def sum_tuple3(num: tuple[float, float, float]) -> float:
    num1, num2, num3 = num
    return num1 + num2 + num3


@transformer
def to_string(num: float) -> str:
    return str(num)


@transformer
def tuple_concatenate(strs: tuple[str, float]) -> str:
    return strs[0] + str(strs[1])


@transformer
def duplicate(string: str) -> str:
    return string + string


@transformer
def natural_logarithm(num: float) -> float:
    if num < 0:
        raise LnOfNegativeNumber(num)
    else:
        return math.log(num, math.e)


@partial_transformer
def logarithm(arg: float, base: float) -> float:
    return math.log(arg) / math.log(base)


@partial_transformer
def format_currency(number: float, thousands_separator: str) -> str:
    return f"{number}:{thousands_separator}.2f"


@partial_transformer
def repeat(content: str, n_times: int, linebreak: bool) -> str:
    repeated = content * n_times
    if linebreak:
        repeated += "\n"
    return repeated



import unittest

from tests.lib.exceptions import (
    NumberIsEven,
    NumberLessThanOrEquals10,
    NumbersNotEqual,
    NumberIsOdd,
)
from tests.lib.ensurers import (
    is_odd,
    same_value,
    is_even,
    is_greater_than_10,
    same_value_int,
)
from gloe import ensure, transformer, partial_transformer


class TestTransformerEnsurer(unittest.TestCase):
    def test_ensure_decorator(self):
        @ensure(incoming=[is_even])
        @transformer
        def divide_by_2(num: int) -> float:
            return num / 2

        self.assertRaises(NumberIsOdd, lambda: divide_by_2(3))
        self.assertEqual(divide_by_2(2), 1)

    def test_output_ensurer(self):
        @ensure(outcome=[is_greater_than_10])
        @transformer
        def multiply_by_2(num: float) -> float:
            return num * 2

        self.assertRaises(NumberLessThanOrEquals10, lambda: multiply_by_2(4))
        self.assertEqual(multiply_by_2(6), 12)

        @ensure(changes=[same_value_int])
        @transformer
        def _plus1(num: int) -> int:
            return num + 1

        self.assertRaises(NumbersNotEqual, lambda: _plus1(4))

        @ensure(outcome=[is_greater_than_10], changes=[same_value])
        @transformer
        def identity(num: float) -> float:
            return num

        @ensure(changes=[same_value], outcome=[is_greater_than_10])
        @transformer
        def identity2(num: float) -> float:
            return num

        self.assertEqual(identity(11), 11)
        self.assertEqual(identity2(11), 11)
        self.assertRaises(NumberLessThanOrEquals10, lambda: identity(9))

    def test_many_ensurers(self):
        @ensure(incoming=[is_even], outcome=[is_greater_than_10])
        @transformer
        def multiply_by_2(num: int) -> float:
            return num * 2

        self.assertRaises(NumberIsOdd, lambda: multiply_by_2(3))
        self.assertRaises(NumberLessThanOrEquals10, lambda: multiply_by_2(4))
        self.assertRaises(NumberIsOdd, lambda: multiply_by_2(7))
        self.assertEqual(multiply_by_2(6), 12)

    def test_ensurer_init_incoming(self):
        @ensure(incoming=[is_even])
        @partial_transformer
        def multiply_by_even(n1: int, n2: int) -> int:
            return n1 * n2

        self.assertRaises(NumberIsOdd, lambda: multiply_by_even(2)(3))
        self.assertEqual(multiply_by_even(2)(4), 8)

    def test_ensurer_init_outcome(self):
        @ensure(incoming=[is_even])
        @partial_transformer
        def multiply_by_equals_even(n1: int, n2: int) -> int:
            return n1 * n2

        self.assertRaises(NumberIsOdd, lambda: multiply_by_equals_even(3)(3))
        self.assertEqual(multiply_by_equals_even(3)(2), 6)

    def test_ensurer_init_incoming_outcome(self):
        @ensure(incoming=[is_even], outcome=[is_even])
        @partial_transformer
        def multiply_by_even_equals_even(n1: int, n2: int) -> int:
            return n1 * n2

        @ensure(incoming=[is_odd], outcome=[is_even])
        @partial_transformer
        def multiply_by_odd_equals_even(n1: int, n2: int) -> int:
            return n1 * n2

        self.assertRaises(NumberIsOdd, lambda: multiply_by_even_equals_even(4)(3))
        self.assertEqual(multiply_by_even_equals_even(4)(6), 24)

        self.assertRaises(NumberIsEven, lambda: multiply_by_odd_equals_even(3)(4))
        self.assertEqual(multiply_by_odd_equals_even(4)(3), 12)
